/******************** (C) COPYRIGHT 2010 Intersil ********************
* File Name          : jpeg_en_driver.c
* Author             : zhang zhao
* Revision           : 1.2.1.1289
* Date               : 2012-04-05
* Description        : the tc_driver's staff for jpeg encode
* Change log         : none
********************************************************************************
* THE PRESENT SOFTWARE IS DISTRIBUTED IN THE HOPE THAT IT WILL PROVIDE CUSTOMERS 
* WITH CODING INFORMATION TO SPEED UP THEIR PRODUCT DEVELOPMENT, BUT WITHOUT
* ANY WARRANTY; WITHOUT EVEN THE IMPLIED WARRANTY OF MERCHANTABILITY OR FITNESS
* FOR A PARTICULAR PURPOSE. AS A RESULT, TECHWELL SHALL NOT BE HELD LIABLE FOR ANY
* DIRECT, INDIRECT OR CONSEQUENTIAL DAMAGES WITH RESPECT TO ANY CLAIMS ARISING
* FROM THE CONTENT OF SUCH SOFTWARE AND/OR THE USE MADE BY CUSTOMERS OF THE
* CODING INFORMATION CONTAINED HEREIN IN CONNECTION WITH THEIR PRODUCTS.
*******************************************************************************/



#include <isil5864/tc_common.h>

#if defined(PCI_PLATFORM)



/*******************************************************************************
* Function Name  : isil_jpeg_driver_init
* Description    : create a driver instance for jpeg according to arg
* Input          : struct isil_chip_device *tcd,struct isil_service_device * tsd,void * para,struct cmd_arg *arg
* Output         : NULL
* Return         : 0:success; !0:error
*******************************************************************************/
int isil_jpeg_driver_init(struct isil_chip_device *tcd, struct isil_service_device *tsd, void * para,struct cmd_arg *arg)
{
    int ret, phy_slot_id;
    unsigned long bus_id,chip_id,lg_id;
    isil_logic_phy_table_t *logic_map_phy;
    isil_chip_t  *chip = (isil_chip_t  *)para;
    isil_jpeg_logic_encode_chan_t	*jed;

    bus_id = get_busid_from_epobj(&tcd->epobj);
    chip_id = get_chipid_from_epobj(&tcd->epobj);
    lg_id = arg->channel_idx;
    logic_map_phy = &chip->chip_driver->logic_map_phy;

    jed = chip->jpeg_encode_chan;
    jed += lg_id;

    if(logic_map_phy->op->find_phy_by_logic(logic_map_phy, LOGIC_MAP_TABLE_BIND_JPEGE, lg_id, &phy_slot_id) < 0){
        dbg("can't find phy_id by logic_id %lu\n", lg_id);
        return -ENODEV;
    }
    ret = init_isil_jpeg_encode_chan(jed, chip->bus_id, chip->chip_id, phy_slot_id, chip);
    if(ret){
        dbg("jpeg driv init failed!\n");
        return ret;
    }
    tsd->ped = &jed->opened_logic_chan_ed;
    return ret;
}


/*******************************************************************************
* Function Name  : isil_jpeg_driver_remove
* Description    : remove a driver instance for jpeg according to arg
* Input          : struct isil_chip_device *tcd,struct isil_service_device * tsd,void * para,struct cmd_arg *arg
* Output         : NULL
* Return         : NULL
*******************************************************************************/
void isil_jpeg_driver_remove(struct isil_chip_device *tcd, struct isil_service_device *tsd, void * para,struct cmd_arg *arg)
{
    isil_jpeg_logic_encode_chan_t *vj;

    //tc_trace;
    if(!tsd->ped){
        dbg("tsd has no ed!\n");
        return;
    }
    vj = container_of(tsd->ped,isil_jpeg_logic_encode_chan_t,opened_logic_chan_ed);
    remove_isil_jpeg_encode_chan(vj);        
}


struct isil_dev_id  isil_jpeg_en_dev_id =
{
    .epobj = {
		.vendor_id = ISIL,
#if	defined(PCI_PLATFORM)
        .bus_id =  ISIL_ED_PCI_BUS,
#else
        .bus_id =  ISIL_ED_HOST_BUS,
#endif
        .chip_id = TW5864,
        .func_id = (STREAM_TYPE_MAJOR << 16) | CODEC_VIDEO_MJPG,
        .type = ISIL_ENCODER,                
    },
    .version = 1,
};


struct tsd_driver_operations  isil_jpeg_en_ops =
{
    .init = isil_jpeg_driver_init,
    .remove = isil_jpeg_driver_remove,
};

struct isil_service_driver   isil_jpeg_encoder_driv =
{
    .driver = {
        .name = "tw5864_jpeg_driv",
    },
    .device_id = {
        .tid = &isil_jpeg_en_dev_id,
        .ops = &isil_jpeg_en_ops,
        //.driver_data =  //parameters
    },
};

/*******************************************************************************
* Function Name  : isil_jpeg_en_driver_init
* Description    : register the jpeg driver related dev_table_id to tc_core
* Input          : NULL
* Output         : NULL
* Return         : NULL
*******************************************************************************/
int isil_jpeg_en_driver_init(void)
{
    //isil_jpeg_encoder_driv.device_id.para = para;
    return register_tc_driver(&isil_jpeg_encoder_driv);
}


/*******************************************************************************
* Function Name  : isil_jpeg_en_driver_remove
* Description    : unregister the jpeg driver related dev_table_id from tc_core
* Input          : NULL
* Output         : NULL
* Return         : NULL
*******************************************************************************/
void isil_jpeg_en_driver_remove()
{
    unregister_tc_driver(&isil_jpeg_encoder_driv);
}
#endif

