/******************** (C) COPYRIGHT 2010 Intersil ********************
* File Name          : tc.c
* Author             : zhang zhao
* Revision           : 1.2.1.1289
* Date               : 2012-04-05
* Description        : the core file for tc_core, if perform a bus liked framework
*                      to allow new tc_devs and tc_drivers to register into it.
* Change log         : none
********************************************************************************
* THE PRESENT SOFTWARE IS DISTRIBUTED IN THE HOPE THAT IT WILL PROVIDE CUSTOMERS 
* WITH CODING INFORMATION TO SPEED UP THEIR PRODUCT DEVELOPMENT, BUT WITHOUT
* ANY WARRANTY; WITHOUT EVEN THE IMPLIED WARRANTY OF MERCHANTABILITY OR FITNESS
* FOR A PARTICULAR PURPOSE. AS A RESULT, TECHWELL SHALL NOT BE HELD LIABLE FOR ANY
* DIRECT, INDIRECT OR CONSEQUENTIAL DAMAGES WITH RESPECT TO ANY CLAIMS ARISING
* FROM THE CONTENT OF SUCH SOFTWARE AND/OR THE USE MADE BY CUSTOMERS OF THE
* CODING INFORMATION CONTAINED HEREIN IN CONNECTION WITH THEIR PRODUCTS.
*******************************************************************************/

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/slab.h>
#include <linux/i2c.h>
#include <linux/fs.h>
#include <linux/cdev.h>
#include <linux/init.h>
#include <linux/idr.h>
#include <linux/seq_file.h>
#include <linux/platform_device.h>
#include <linux/mutex.h>
#include <linux/completion.h>
#include <asm/uaccess.h>
#include <linux/kthread.h>
#include <linux/unistd.h>
#include <asm/unistd.h>
#include <linux/unistd.h>
#include <linux/fcntl.h>
#include <linux/namei.h>
#include <linux/radix-tree.h>
#include <linux/sysfs.h>
#include <asm/current.h>
#include <isil5864/tc_common.h>
//#include <linux/kprobes.h>

MODULE_LICENSE("GPL");

static struct kobject * kroot = NULL;
static LIST_HEAD(isil_chipdev);
static DEFINE_MUTEX(mutex_chipdev);

static LIST_HEAD(isil_chipdrv);
static DEFINE_MUTEX(mutex_chipdrv);

static LIST_HEAD(isil_tc_dev);
static DEFINE_MUTEX(mutex_tc_dev);

static LIST_HEAD(isil_tc_drv);
static DEFINE_MUTEX(mutex_tc_drv);

//static RAW_NOTIFIER_HEAD(isil_chipdev_chain);
//struct class *my_class;
//static struct radix_tree_root look_tb = RADIX_TREE_INIT(GFP_KERNEL);



static int init_nfm_bitmap(void);
static void free_nfm_bitmap(void);
static int get_free_minor(struct cmd_arg *arg);
static void put_free_minor(unsigned long minor,unsigned long type);
static void  make_service_key(epkey_t *key,struct isil_chip_device * tcd, struct cmd_arg * arg);
//static void isil_dir_release(struct kobject * kobj);
static struct kobject * isil_create_dir(char * name, struct kobject *parent);
static void isil_remove_dir(struct kobject * kobj);
static int bind_chip_driver(struct isil_chip_device * tcd);
static void unbind_chip_driver(struct isil_chip_device * tcd);

/*
static void chip_dev_get(struct isil_chip_device * tcd);
static void chip_dev_put(struct isil_chip_device * tcd);
static void service_dev_get(struct isil_service_device * tsd);
static void service_dev_put(struct isil_service_device * tsd);
*/
static int  get_dir(struct isil_chip_device *tcd, unsigned long type, struct kobject **dir);
//static struct tsd_file_operations * lookup_tsd_dev_fops_from_dev(struct isil_service_device *tsd);
static int request_devno_resource(void);
static void free_devno_resource(void);

struct tc_hb_pool *thb_pool = NULL;
char *algorithm[] = {
	"h264","mjpg","alaw","ulaw","abadpcm",
	"mpmlq","adpcm","ldcecp","csacelp",
};
char *stream[] ={"major","minor","pic","mv","audio",};
char *tc_dev_name[] = {"isil_chip","isil_encoder","isil_decoder","isil_video_in","isil_video_out","isil_audio_in","isil_audio_out",};





static LIST_HEAD(cl_head);
static unsigned long resource_major[MAX_SERVICE_NR];

/*this indicate the minor numbers which we can use, 0 is free, 1 is using*/
static unsigned long *nfm_bitmap[MAX_SERVICE_NR];


void lock_chipdev(void)
{
    mutex_lock(&mutex_chipdev);
}


void unlock_chipdev(void)
{
    mutex_unlock(&mutex_chipdev);
}


struct list_head * get_chipdev_listhead(void)
{
    return &isil_chipdev;
}


/*******************************************************************************
* Function Name  : tc_get_inode_from_file
* Description    : get inode from file
* Input          : struct file *f
* Output         : NULL
* Return         : pointer which point to inode, should be !NULL
*******************************************************************************/
struct inode * tc_get_inode_from_file(struct file *f)
{
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 20)
    return f->f_dentry->d_inode;
#else
    return f->f_path.dentry->d_inode;
#endif

}







/*******************************************************************************
* Function Name  : init_nfm_bitmap
* Description    : allocate the memory this bitmap used
* Input          : NULL
* Output         : NULL
* Return         : 0:success; -1:failed
*******************************************************************************/
static int init_nfm_bitmap(void)
{
    int idx;
    unsigned long len = tc_round_up(MAX_MINOR_NR,8);


    for(idx=0; idx<MAX_SERVICE_NR; idx++)
    {
        nfm_bitmap[idx] = kzalloc(len,GFP_KERNEL);
        if(!nfm_bitmap[idx])  goto clean;
    }
    return 0;
clean:
    dbg("in %s failed to get bitmap space!\n",__FUNCTION__);
    for(;idx > 0; idx--)
        kfree(nfm_bitmap[idx-1]);
    return -1;
}

/*******************************************************************************
* Function Name  : free_nfm_bitmap
* Description    : free the memory this bitmap used
* Input          : NULL
* Output         : NULL
* Return         : NULL
*******************************************************************************/
static void free_nfm_bitmap(void)
{
    unsigned long i;

    for(i=0; i<MAX_SERVICE_NR; i++)
        kfree(nfm_bitmap[i]);

}


/*******************************************************************************
* Function Name  : get_free_minor
* Description    : get the free minor bumber to use
* Input          : struct cmd_arg *arg
* Output         : NULL
* Return         : >=0 && <=255: success; -1:failed
*******************************************************************************/
static int get_free_minor(struct cmd_arg *arg)
{
    int ret = MAX_MINOR_NR;
    int idx = arg->type;

    ret = find_first_zero_bit(nfm_bitmap[idx],MAX_MINOR_NR);
    if(ret >= MAX_MINOR_NR){
        dbg("there is no free minor!\n");
        return -1;
    }
    //dbg("the first zero bit is %d\n",ret);
    set_bit(ret,nfm_bitmap[idx]);
    return ret;
}

/*******************************************************************************
* Function Name  : put_free_minor
* Description    : return the minor number to free, other guys can use
* Input          : unsigned long minor,unsigned long type
* Output         : NULL
* Return         : NULL
*******************************************************************************/
static void put_free_minor(unsigned long minor,unsigned long type)
{	
    clear_bit(minor,nfm_bitmap[type]);
}


/*******************************************************************************
* Function Name  : tc_major_2_type
* Description    : translate major to an type index 
* Input          : unsigned long major
* Output         : NULL
* Return         : >=0:type; -1:failed
*******************************************************************************/
static int tc_major_2_type(unsigned long major)
{
    int i;

    for(i = 0; i < MAX_SERVICE_NR; i++){
        if(resource_major[i] == major)
        	return i;
    }
    return -1;

}


/*******************************************************************************
* Function Name  : make_service_key
* Description    : make up a key from given parameters
* Input          : epkey_t *key ,struct isil_chip_device * tcd, struct cmd_arg * arg
* Output         : NULL
* Return         : NULL
*******************************************************************************/
static void make_service_key(epkey_t *key ,struct isil_chip_device * tcd, struct cmd_arg * arg)
{
    int bus_id = get_bus_id(tcd->epobj.key);
    int chip_id = tcd->minor;
    int pid = 0;
    int lid = arg->channel_idx;	

    make_key(key,arg->stream,arg->algorithm,bus_id,chip_id, pid, lid);
}

/*******************************************************************************
* Function Name  : isil_dir_release
* Description    : the release callback of this kobj of dir
* Input          : struct kobject * kobj
* Output         : NULL
* Return         : NULL
*******************************************************************************/
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 25)  
static void isil_dir_release(struct kobject * kobj)
{
    kfree(kobj);
}

static struct kobj_type  isil_dir = {
    .release        = isil_dir_release,
    .sysfs_ops      = NULL,
    .default_attrs  = NULL,
}; 
#endif




/*******************************************************************************
* Function Name  : isil_create_dir
* Description    : create the dir with given name and parent
* Input          : char * name, struct kobject *parent
* Output         : NULL
* Return         : !NULL:the kobj of new dir; NULL: failed
*******************************************************************************/
static struct kobject * isil_create_dir(char * name, struct kobject *parent)
{
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 25)
    int ret;     
    struct kobject* kobj = kzalloc(sizeof(struct kobject),GFP_ATOMIC);

    if(!kobj)
        return NULL;

    kobj->parent = parent;
    kobj->ktype = &isil_dir;
    kobject_init(kobj);
    kobject_set_name(kobj, "%s", name);
    ret = kobject_register(kobj);
    if(ret){
        kobject_del(kobj);
        return NULL;
    }
    else
    {
        return kobj;
    }
#else

    return kobject_create_and_add(name,parent);
#endif
}


/*******************************************************************************
* Function Name  : isil_remove_dir
* Description    : remove the dir form one kobj
* Input          : struct kobject * kobj
* Output         : NULL
* Return         : NULL
*******************************************************************************/
static void isil_remove_dir(struct kobject * kobj)
{
    if(!kobj)
        return;

    //tc_trace;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 25)
    kobject_unregister(kobj);
#else
    kobject_put(kobj);
#endif
} 



static int isil_device_match(struct device *dev, struct device_driver *drv)
{
    return 0;
}


static ssize_t chip_name_show(struct device *dev, struct device_attribute *attr, char *buf)
{
    return 0;
}


#if LINUX_VERSION_CODE >= KERNEL_VERSION(3,12,0)
static DEVICE_ATTR_RO(chip_name);
static struct attribute *isil_dev_attrs[] = {
    &dev_attr_chip_name.attr,
    NULL,
};
ATTRIBUTE_GROUPS(isil_dev);
#else
static struct device_attribute isil_dev_attrs[] = {
    __ATTR(name, S_IRUGO, chip_name_show, NULL),
    { },
};
#endif

static struct bus_type isil_virtual_bus = {
    .name		= "isil_virtual_bus",
#if LINUX_VERSION_CODE >= KERNEL_VERSION(3,12,0)
    .dev_groups = isil_dev_groups,
#else
    .dev_attrs	= isil_dev_attrs,
#endif
    .match		= isil_device_match,
};


/*******************************************************************************
* Function Name  : find_chip_dev
* Description    : find chip device form minor
* Input          : int minor
* Output         : NULL
* Return         : !NULL: success; NULL: failed
*******************************************************************************/
struct isil_chip_device * find_chip_dev(int minor)
{
    struct list_head * p = &isil_chipdev;
    struct isil_chip_device *tcd = NULL;

    if((p->next == p->prev) && (p->next == &isil_chipdev))
    {
        dbg("out from here? and minor=%d\n",minor);
        return NULL;
    }
    mutex_lock(&mutex_chipdev);
    p = p->next;
    for(; p != &isil_chipdev; p = p->next)
    {
        tcd = container_of(p,struct isil_chip_device, list);
        if(tcd->minor == minor){
            mutex_unlock(&mutex_chipdev);
            return tcd;
        }
    }

    mutex_unlock(&mutex_chipdev);
    return NULL;
}


/*******************************************************************************
* Function Name  : twc_release
* Description    : open support function in chip device
* Input          : struct inode *inode, struct file *filp
* Output         : NULL
* Return         : 0:success; !0: failed
*******************************************************************************/
int twc_open(struct inode *inode, struct file *file)
{
    unsigned int minor = iminor(inode);
    int ret = 0;
    struct isil_chip_device *tcd;

    if(minor >= 256)
        return -ENODEV;

    tcd = find_chip_dev(minor);
    if(tcd == NULL)
        return -ENODEV;
	
    if(tcd->tcd_fops && tcd->tcd_fops->open)
        ret = tcd->tcd_fops->open(tcd,file);

    return ret;
}


/*******************************************************************************
* Function Name  : twc_release
* Description    : release support function in chip device
* Input          : struct inode *inode, struct file *filp
* Output         : NULL
* Return         : 0:success; !0: failed
*******************************************************************************/
int twc_release(struct inode *inode, struct file *filp)
{
    unsigned int minor = iminor(inode);
    int ret = 0;
    struct isil_chip_device *tcd;

    if(minor >= 256)
        return -ENODEV;

    tcd = find_chip_dev(minor);
    if(!tcd)
        return -ENODEV;
    
    if(tcd->tcd_fops && tcd->tcd_fops->release)
        tcd->tcd_fops->release(tcd,filp);

    return ret;	
}

/*******************************************************************************
* Function Name  : twc_read
* Description    : write support function in chip device
* Input          : struct file *filp, const char *buff, size_t count, loff_t *offp
* Output         : NULL
* Return         : >=0: number of byte read ; <0: failed
*******************************************************************************/
ssize_t twc_read(struct file *filp, char *buff, size_t count, loff_t *offp)
{
    ssize_t ret;
    struct isil_chip_device * tcd;

    tcd = find_chip_dev(iminor(tc_get_inode_from_file(filp)));
    if(!tcd)
        return -ENODEV;
    if(!tcd->tcd_fops->read)
        return -EINVAL;

    ret = tcd->tcd_fops->read(filp,buff,count,offp);
    return ret;
}


/*******************************************************************************
* Function Name  : twc_write
* Description    : write support function in chip device
* Input          : struct file *filp, const char *buff, size_t count, loff_t *offp
* Output         : NULL
* Return         : >=0: number of byte wrote ; <0: failed
*******************************************************************************/
ssize_t twc_write(struct file *filp,  const char __user *buff, size_t count, loff_t *offp)
{
    ssize_t ret;
    struct isil_chip_device * tcd;

    tcd = find_chip_dev(iminor(tc_get_inode_from_file(filp)));
    if(!tcd)
        return -ENODEV;

    if(!tcd->tcd_fops->write)
        return -EINVAL;

    ret = tcd->tcd_fops->write(filp,buff,count,offp);
    return ret;
}

/*******************************************************************************
* Function Name  : twc_ioctl
* Description    : ioctl support function in chip device
* Input          : struct file * filp,unsigned int cmd, unsigned long arg
* Output         : NULL
* Return         : 0: success; !0: failed
*******************************************************************************/
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 36)
long twc_ioctl(struct file * filp,unsigned int cmd, unsigned long arg)
#else
int twc_ioctl(struct inode * inode,struct file * filp,unsigned int cmd, unsigned long arg)
#endif
{
    int ret = 0;
    struct isil_chip_device *tcd;
    struct cmd_arg cmdarg;
    struct cmd_arg *cmdarg2;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 36)
    tcd = find_chip_dev(iminor(filp->f_path.dentry->d_inode));
#else
    tcd = find_chip_dev(iminor(inode));
#endif

    if(!tcd){
        ISIL_DBG(ISIL_DBG_INFO, ": can not find the tcd by minor number!!!!!!\n");
        return -ENODEV;
    }

    if(!tcd->tcd_fops  || !tcd->tcd_fops->ioctl){
        ISIL_DBG(ISIL_DBG_INFO, ": you have not ioctl func registered!\n");
        return -EINVAL;
    }
    switch(cmd)
    {
    case ISIL_CHIP_CREATE_CHAN:
	    if (copy_from_user(&cmdarg, (void *)arg, sizeof(struct cmd_arg)) != 0) {
		    printk("failed to copy from user\n");
	    }
	    ret = create_service_device(tcd, (void *) &cmdarg, 2);
	    cmdarg2 = (struct cmd_arg *) arg;
	    if (copy_to_user((void *)&cmdarg2->dev_info2,(void *)&cmdarg.dev_info2,sizeof(struct device_info)) != 0) {
		    printk("failed to copy from user\n");
	    }
	    break;
    case ISIL_CHIP_RELEASE_CHAN:
	    if (copy_from_user(&cmdarg, (void *)arg, sizeof(struct cmd_arg)) != 0) {
		    printk("failed to copy from user\n");
	    }
	    release_service_device(tcd,(void *)&cmdarg);
	    break;
    default:
            ret = tcd->tcd_fops->ioctl(filp,cmd,arg);
            break;
    }
    return ret;
}


#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 36)
long twc_compat_ioctl(struct file * filp,unsigned int cmd, unsigned long arg)
#else
int twc_compat_ioctl(struct inode * inode,struct file * filp,unsigned int cmd, unsigned long arg)
#endif
{
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 36)
	return twc_ioctl(filp, cmd, arg);
#else
	return twc_ioctl(inode, filp, cmd, arg);
#endif	
}
	

struct file_operations isil_fops = {
    .owner = THIS_MODULE,
    .open = twc_open,
    .release = twc_release,
    .read = twc_read,
    .write = twc_write,
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 36)
    .unlocked_ioctl = twc_ioctl,
#else
    .ioctl = twc_ioctl,
#endif
    .compat_ioctl = twc_compat_ioctl,    
};








/*******************************************************************************
* Function Name  : find_service_dev
* Description    : find service device form major and minor
* Input          : int major,int minor
* Output         : NULL
* Return         : !NULL: success; NULL: failed
*******************************************************************************/
struct isil_service_device * find_service_dev(int major,int minor)
{
    struct list_head  *p = &isil_chipdev;
    struct list_head  *q = NULL;
    int type; //= major-ISIL_SERVICE_BASE_MAJOR;
    struct isil_chip_device * tcd = NULL;
    struct isil_service_device *tsd = NULL;


    if((p->next == p->prev) && (p->next == &isil_chipdev))
    {
        dbg("nothing in device list!\n");
        return NULL;
    }

    type = tc_major_2_type(major);
    if(type <= 0 || type > MAX_SERVICE_NR - 1){
    	dbg("wrong major in %s\n", __FUNCTION__);
    	return NULL;
    }
    type = type - 1;//decrease the major index, because of the chip dev
    mutex_lock(&mutex_chipdev);
    p = p->next;
    for(; p != &isil_chipdev; p = p->next){
        tcd = container_of(p,struct isil_chip_device, list);
        if((q = &tcd->tsd_list[type]) == NULL)	
            goto error;

        q = q->next;
        for(; q != &tcd->tsd_list[type]; q = q->next){
            tsd = container_of(q,struct isil_service_device,list);
            if(tsd->minor == minor)
                goto find;
        }
    }

find:
    mutex_unlock(&mutex_chipdev);
    return tsd;
error:
    dbg("BUG! the listhead of given type is NULL!\n");
    mutex_unlock(&mutex_chipdev);
    return NULL;

}


/*******************************************************************************
* Function Name  : tws_open
* Description    : open support function in service device
* Input          : struct inode *inode, struct file *file
* Output         : NULL
* Return         : 0: success; !0: failed
*******************************************************************************/
int tws_open(struct inode *inode, struct file *file)
{
    int major = imajor(inode);
    int minor = iminor(inode);
    int ret = 0;
    struct isil_service_device *tsd = NULL;

    ISIL_DBG(ISIL_DBG_INFO, ": major:[%d]; minor:[%d]\n",major,minor);
    tsd = find_service_dev(major,minor);
    if(tsd == NULL)
        return -ENODEV;

	if(!(tsd->tsd_fops)){
            dbg("you have no tsd_fops registered!\n");
            return -EINVAL;
        }
        if(!tsd->tsd_fops->open){
            dbg("BUG: have no open function");
            return -EINVAL;
        }

        dbg("major[%d],minor[%d]\n",major,minor);
        ret = tsd->tsd_fops->open(tsd,file);

        return ret;
}


/*******************************************************************************
* Function Name  : tws_release
* Description    : release support function in service device
* Input          : struct inode *inode, struct file *filp
* Output         : NULL
* Return         : 0: success ; !0: failed
*******************************************************************************/
int tws_release(struct inode *inode, struct file *filp)
{
    int ret = 0;
    int major = imajor(inode);
    int minor = iminor(inode);
    struct isil_service_device *tsd = NULL;

    ISIL_DBG(ISIL_DBG_INFO, ": major:[%d]; minor:[%d]\n",major,minor);
    tsd = find_service_dev(major,minor);
    if(!tsd){
        ISIL_DBG(ISIL_DBG_INFO, ": can not find the tsd!!!!!!!!\n");
        return -ENODEV;
    }

    if(!tsd->tsd_fops || !tsd->tsd_fops->release){
        dbg("no tsd_fops or release\n");
        return 0;
    }
    ret = tsd->tsd_fops->release(filp);
    return ret;	
}

/*******************************************************************************
* Function Name  : tws_read
* Description    : read support function in service device
* Input          : struct file *filp, const char *buff, size_t count, loff_t *offp
* Output         : NULL
* Return         : >=0: number of byte readed ; <0: failed
*******************************************************************************/
ssize_t tws_read(struct file *filp, char *buff, size_t count, loff_t *offp)
{
    int ret = 0;
    struct isil_service_device * tsd = (struct isil_service_device *)filp->private_data; /*the tsd goes first in different dev structs*/	

    if(filp->private_data == NULL){
        dbg("error: you get call %s without opened!\n",__FUNCTION__);
        return -EINVAL;
    }

    if(NULL == tsd->tsd_fops){
        dbg("no tsd fop registered!\n");
        return -EINVAL;
    }

    if(tsd->tsd_fops->read)
        ret = tsd->tsd_fops->read(filp,buff,count,offp);					
    return ret;
}

/*******************************************************************************
* Function Name  : tws_write
* Description    : write support function in service device
* Input          : struct file *filp, const char *buff, size_t count, loff_t *offp
* Output         : NULL
* Return         : >=0: number of byte wrote ; <0: failed
*******************************************************************************/
ssize_t tws_write(struct file *filp, const char *buff, size_t count, loff_t *offp)
{
    int ret = 0;
    struct isil_service_device * tsd = (struct isil_service_device *)filp->private_data; /*the tsd goes first in different dev structs*/	

    if(filp->private_data == NULL){
        dbg("error: you get call %s without opened!\n",__FUNCTION__);
        return -EINVAL;
    }

    if(NULL == tsd->tsd_fops){
        dbg("no tsd fop registered!\n");
        return -EINVAL;
    }

    if(tsd->tsd_fops->write)
        ret = tsd->tsd_fops->write(filp,buff,count,offp);					


    return ret;


}


/*******************************************************************************
* Function Name  : tws_ioctl
* Description    : ioctl support function in service device
* Input          : struct file * filp,unsigned int cmd, unsigned long arg
* Output         : NULL
* Return         : 0: success; !0: failed
*******************************************************************************/
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 36)
long tws_ioctl(struct file * filp,unsigned int cmd, unsigned long arg)
#else
int tws_ioctl(struct inode * inode,struct file * filp,unsigned int cmd, unsigned long arg)
#endif
{
    int ret; 
    struct isil_service_device * tsd = (struct isil_service_device *)filp->private_data;	

    if(filp->private_data == NULL){
        dbg("call %s without opened!\n",__FUNCTION__);
        return -EINVAL;
    }
    if(!tsd->tsd_fops || !tsd->tsd_fops->ioctl){
        dbg("no tsd fop registered!\n");
        return -EINVAL;
    }
    ret = tsd->tsd_fops->ioctl(filp,cmd,arg);
    return ret;
}


/*******************************************************************************
* Function Name  : tws_poll
* Description    : poll support function in service device
* Input          : struct file *file, struct poll_table_struct *wait
* Output         : NULL
* Return         : 0: success; !0: failed
*******************************************************************************/
unsigned int tws_poll(struct file *file, struct poll_table_struct *wait)
{
    unsigned int ret = 0;
    struct isil_service_device * tsd = (struct isil_service_device *)file->private_data; /*the tsd goes first in different dev structs*/	

    if(NULL == file->private_data){
        dbg("call %s without opened!\n",__FUNCTION__);
        return 0;
    }
    if(NULL == tsd->tsd_fops){
        dbg("no tsd fop registered!\n");
        return 0;
    }

    if(!tsd->tsd_fops->poll){
        ISIL_DBG(ISIL_DBG_INFO, ": no poll function!\n");
        return ret;
    }
    ret = tsd->tsd_fops->poll(file,wait);
    return ret;
}






#if 0
void tc_vm_open(struct vm_area_struct *vma)
{
    //struct tc_mapping *map = vma->vm_private_data;

    dbg("vm_open virtual:%X phy:%d\n",vma->vm_start,vma->vm_pgoff << PAGE_SHIFT);//%p [count=%d,vma=%08lx-%08lx]\n",map,map->count,vma->vm_start,vma->vm_end);
    //map->count++;
}

void tc_vm_close(struct vm_area_struct *vma)
{
    dbg("vm_close \n");//[count=%d,vma=%08lx-%08lx]\n",map,map->count,vma->vm_start,vma->vm_end);
    return;
}


struct page * tc_vm_nopage(struct vm_area_struct *vma, unsigned long vaddr,int *type)
{
    struct page *page = NOPAGE_SIGBUS;//NOPAGE_SIGBUS=NULL
    struct isil_audio_en *tae = vma->vm_private_data;
    unsigned long off = 0;
    struct tc_queue *q = tae->queue;
    struct tc_buffer_kern * tck = q->bufs[q->rp];

    dbg("nopage: fault @ %08lx [vma %08lx-%08lx]\n",vaddr,vma->vm_start,vma->vm_end);
    if (vaddr > vma->vm_end)
        return NOPAGE_SIGBUS;

    /* get pages(the low level driver handled specified) here!*/	
    off = (vaddr - vma->vm_start) + (vma->vm_pgoff << PAGE_SIZE);
    if(off > 4096*3){
        dbg("%s---------> out og range!\n",__FUNCTION__);
        return page;
    }

    dbg("%s--->off = %d\n",__FUNCTION__,off);
    mutex_lock(&tae->queue->lock);
    page = virt_to_page(tck->start);
    dbg("give the %d tck ;tck->off_to_page = %d\n",q->rp,tck->off_to_page);
    mutex_unlock(&tae->queue->lock);

    dbg("\n tck's page = %X;tc->start = %X and tck->off_to_page = %d\n",page,tck->start,tck->off_to_page);
    if(!page)
        return NOPAGE_OOM;
    get_page(page);
    if (type)
        *type = VM_FAULT_MINOR;
    return page;
}
struct vm_operations_struct tc_vm_ops =
{
    .open     = tc_vm_open,
    .close    = tc_vm_close,
    .nopage   = NULL,//tc_vm_nopage,
};

#endif


/*******************************************************************************
* Function Name  : mmap_scatter
* Description    : memory map support function for each tc_service device
* Input          : struct vm_area_struct *vma,struct tc_queue *tq, unsigned long size
* Output         : NULL
* Return         : 0: success; !0: failed
*******************************************************************************/
int mmap_scatter(struct vm_area_struct *vma,struct tc_queue *tq, unsigned long size)
{
    int ret=0;
    int idx = 0,cix = 0;
    unsigned long start,pfn;
    unsigned long cn = 0;
    struct tc_buffer_kern *tck;

    start = vma->vm_start;
    tck = tq->bufs[0];

    /*first map header_buf*/
    if(!tck->thb || !tck->thb->data){
        dbg("tck has no header buf or thb->data is NULL\n");
        return -EINVAL;
    }
    pfn = (__pa(tck->thb->data)) >> PAGE_SHIFT;
    if(remap_pfn_range(vma,start,pfn,PAGE_SIZE,PAGE_SHARED)){
        dbg("remap header_buf failed\n");
        return -ENXIO;
    }
    start += PAGE_SIZE;


    /*then net payload...*/
    while((tck = tq->bufs[cix]) && (STATE_READY == get_tck_state(tck))){
        cn = tc_round_up(tck->size,PAGE_SIZE);
        set_tck_state(tck,STATE_BUSY);
        for(idx = 0; idx < cn; idx++){
            if(!tck->sg[idx].carrier || !tck->sg[idx].addr || !tck->sg[idx].len)
                break;
            pfn = (__pa(tck->sg[idx].addr)) >> PAGE_SHIFT;
            if(remap_pfn_range(vma,start,pfn,PAGE_SIZE,PAGE_SHARED)){
                dbg("remap page range failed\n");
                return -ENXIO;
            }			
            start += PAGE_SIZE;
        }
        cix++;
        if(cix >= tq->buf_nr)
            break;
    }
    return ret;
}

/*******************************************************************************
* Function Name  : mmap_single
* Description    : memory map support function for each tc_service device
* Input          : struct vm_area_struct *vma,struct tc_queue *tq
* Output         : NULL
* Return         : 0: success; !0: failed
*******************************************************************************/
int mmap_single(struct vm_area_struct *vma,struct tc_queue *tq)
{
    unsigned long pfn,last_pfn = 0,start;
    struct tc_buffer_kern *tck = tq->bufs[0];
    int idx = 0;

    start = vma->vm_start;

    if(!tck->thb || !tck->thb->data){
        dbg("tck has no header buf or thb->data is NULL\n");
        return -EINVAL;
    }
    /*first map header_buf*/
    pfn = (__pa(tck->thb->data)) >> PAGE_SHIFT;
    if(remap_pfn_range(vma,start,pfn,PAGE_SIZE,PAGE_SHARED)){
        dbg("remap header_buf failed\n");
        return -ENXIO;
    }
    start += PAGE_SIZE;        

    while((tck = tq->bufs[idx])  && (STATE_READY == get_tck_state(tck))){
        pfn = (__pa(tck->start)) >> PAGE_SHIFT;
        set_tck_state(tck,STATE_BUSY);
        if(last_pfn != pfn){
            if (remap_pfn_range(vma,start,pfn,PAGE_SIZE,PAGE_SHARED)){
                dbg("remap page range failed\n");
                return -ENXIO;
            }
            start += PAGE_SIZE;
        }
        last_pfn = pfn;
        idx++;
        if(idx >= tq->buf_nr)
            break;
    }
    return 0;	
}

/*******************************************************************************
* Function Name  : tws_mmap
* Description    : memory map support function for each tc_service device
* Input          : struct file *file, struct vm_area_struct *vma
* Output         : NULL
* Return         : 0: success; !0: failed
*******************************************************************************/
int tws_mmap(struct file *file, struct vm_area_struct *vma)
{
    int ret = 0;
    struct isil_service_device *tsd = (struct isil_service_device *)file->private_data;
    unsigned long offset = vma->vm_pgoff << PAGE_SHIFT;
    unsigned long size = vma->vm_end - vma->vm_start;
    struct tc_queue *q = tsd->queue;

#ifdef STATIC_COUNTING
    get_ic_map_start(tsd);
#endif
    if (offset & ~PAGE_MASK){
        dbg("offset not aligned: %ld\n", offset);
        return -ENXIO;
    }

    if ((vma->vm_flags & VM_WRITE) && !(vma->vm_flags & VM_SHARED)){
        dbg("writeable mappings must be shared, rejecting\n");
        return(-EINVAL);
    }

    if(!q){
        dbg("tsd no queue\n");
        return -EINVAL;
    }


    if(q->unit_size >PAGE_SIZE)
        ret = mmap_scatter(vma,q,size);
    else
        ret = mmap_single(vma,q);

#ifdef STATIC_COUNTING
    calc_ic_map_time(tsd);
#endif
    return ret;
}


struct file_operations tws_fops = {
    .owner = THIS_MODULE,
    .open = tws_open,
    .release = tws_release,
    .read = tws_read,
    .write = tws_write,
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 36)
    .unlocked_ioctl = tws_ioctl,
#else
    .ioctl = tws_ioctl,
#endif
    .mmap  = tws_mmap,
    .poll = tws_poll,
};




void dumy_release(struct device * dev)
{
    return;
}



/*******************************************************************************
* Function Name  : tc_chipdev_alloc
* Description    : Allocates a struct tcd with private data area for driver use
*                : and perform basic initialization.
* Input          : unsigned long priv_size
* Output         : NULL
* Return         : !NULL: success; NULL: failed
*******************************************************************************/
struct isil_chip_device * tc_chipdev_alloc(unsigned long priv_size)
{
    void * p;
    unsigned long size;
    struct isil_chip_device *tcd;

    size = (sizeof(*tcd) + CHIPDEV_ALIGN_MASK) & ~CHIPDEV_ALIGN_MASK;
    size += priv_size + CHIPDEV_ALIGN_MASK;
    size += (PAGE_SIZE-1);
    size >>= 12;
    size <<= 12;
    p = (void *)__get_free_pages(GFP_KERNEL, get_order(size));//kzalloc(size, GFP_KERNEL);
    if(!p){
        ISIL_DBG(ISIL_DBG_INFO, ": no mem!, %lu\n", size);
        return NULL;
    }
    memset(p, 0, size);
    tcd = (struct isil_chip_device *)(((long)p + CHIPDEV_ALIGN_MASK) & ~CHIPDEV_ALIGN_MASK);
    tcd->padded = (char *)tcd - (char *)p;
    tcd->priv_len = priv_size;
    tcd->total_len = size;
    if(priv_size)
        tcd->priv = tcd_priv(tcd);
    return tcd;
}

/*******************************************************************************
* Function Name  : tc_chipdev_free
* Description    : check chip_dev's information
* Input          : struct isil_chip_device * tcd
* Output         : NULL
* Return         : 0: success; -1: failed
*******************************************************************************/
void tc_chipdev_free(struct isil_chip_device * tcd)
{
    char *p = (char *)tcd;

    p = p- tcd->padded;
    free_pages((unsigned long)p,get_order(tcd->total_len));
}





/*******************************************************************************
* Function Name  : check_chipdev_info
* Description    : check chip_dev's information
* Input          : struct isil_chip_device * tcd
* Output         : NULL
* Return         : 0: success; -1: failed
*******************************************************************************/
int check_chipdev_info(struct isil_chip_device * tcd)
{
    if(!tcd->name[0]){
        dbg("no name for chip device!\n");
        return -1;
    }
    return 0;
}



int setup_chip_dev(struct isil_chip_device * tcd)
{
    return 0;
}



/*******************************************************************************
* Function Name  : bind_chip_driver
* Description    : bind one chip driver to the matched tc_chip_device
* Input          : struct isil_chip_device * tcd
* Output         : NULL
* Return         : 0: success; 1: failed
*******************************************************************************/
static int bind_chip_driver(struct isil_chip_device * tcd)
{
    struct list_head *p;
    struct isil_chip_driver *tdr;
    struct isil_dev_table_id *table_id;

    /*walk the list to find the crosponding devices*/
    list_for_each(p,&isil_chipdrv){
        tdr = list_entry(p,struct isil_chip_driver,list);
        table_id = &tdr->device_id;
        if(!table_id->tid){
            ISIL_DBG(ISIL_DBG_INFO, ": BUG,no tid registered with isil_chip_driver!\n");
            return -EINVAL;
        }
        if(chip_device_match(&tcd->epobj,&table_id->tid->epobj)){
            tcd->driver = tdr;
            if(tdr->attach_chip)
                return tdr->attach_chip(tdr,tcd);//driver init	
        }
    }
    return 0;
}


/*******************************************************************************
* Function Name  : unbind_chip_driver
* Description    : unbind chip driver from one chip device.
* Input          : struct isil_chip_device *tcd
* Output         : NULL
* Return         : NULL
*******************************************************************************/
static void unbind_chip_driver(struct isil_chip_device *tcd)
{
    struct list_head *p;
    struct isil_chip_driver *tdr;
    struct isil_dev_table_id *table_id;

    /*walk the list to find the crosponding devices*/
    list_for_each(p,&isil_chipdrv){
        tdr = list_entry(p,struct isil_chip_driver,list);
        table_id = &tdr->device_id;
        if(!table_id->tid){
            ISIL_DBG(ISIL_DBG_INFO, ": BUG, no tid registered with isil_chip_driver!\n");
            return;
        }
        if(chip_device_match(&tcd->epobj,&table_id->tid->epobj)){
            tcd->driver = tdr;
            if(tdr->detach_chip)
                tdr->detach_chip(tdr,tcd);	
        }
    }
}




/*******************************************************************************
* Function Name  : chip_device_match
* Description    : identity two chip device from it's key information epobj_t
* Input          : struct epobj_t *a, struct epobj_t *b
* Output         : NULL
* Return         : 1: matched; 0: not matched
*******************************************************************************/
int chip_device_match(struct epobj_t *a, struct epobj_t *b)
{
	//dbg("a->vid[%lu],a->bid[%lu],a->cid[%lu]---\n b->vid[%lu],b->bid[%lu],b->cid[%lu]---\n",a->vendor_id,
	//	a->bus_id,a->chip_id,b->vendor_id,b->bus_id,b->chip_id);
	if(a->vendor_id == b->vendor_id && a->bus_id == b->bus_id && a->chip_id == b->chip_id)
		return 1;
	else
		return 0;
}


/*note:the tcd->name must be filled before call this function*/
/*******************************************************************************
* Function Name  : register_chip_device
* Description    : register one chip device into tc_core
* Input          : struct isil_chip_device * tcd
* Output         : NULL
* Return         : 0: true; !0: failed
*******************************************************************************/
int register_chip_device(struct isil_chip_device *tcd)
{
    int ret = 0;
    int idx = 0; 
    int minor = -1;
    unsigned long flag = 0;
    char chip_dir_name[50]={0};
    struct kobject *kobj = NULL;
    unsigned long fix_id;
    struct cmd_arg ca = {0};


    if(check_chipdev_info(tcd))
        return -EINVAL;

    ca.type = 0;
    minor = get_free_minor(&ca);
    if(minor < 0 || minor > 256)
    {
        dbg("minor is %d",minor);
        return -EINVAL;
    }

    snprintf(chip_dir_name,sizeof(chip_dir_name),"%s",tcd->name);
    kobj = isil_create_dir(chip_dir_name,kroot);
    if(kobj == NULL){
        dbg("create dir failed!\n");
        return -EINVAL;
    }

    tcd->cdev = (struct cdev *)cdev_alloc();
    if(tcd->cdev == NULL){
        dbg("cdev_alloc failed!\n");
        ret = -ENOMEM;
        goto cleanup;
    }

    cdev_init(tcd->cdev,&isil_fops);
    tcd->cdev->owner = THIS_MODULE;//tcd->tcd_fops->owner;
    ret = cdev_add(tcd->cdev,MKDEV(resource_major[ISIL_CHIP],minor),1);
    if(ret < 0){
        flag = 1;
        dbg("cdev_add failed!\n");
        goto cleanup;
    }

    for(idx = 0; idx < sizeof(tcd->tsd_list)/sizeof(struct list_head) ; idx++){
        INIT_LIST_HEAD(&tcd->tsd_list[idx]);
        tcd->dirs[idx] = NULL;
    }
    tcd->minor = minor;

    tcd->dev.parent = NULL;
    tcd->dev.kobj.parent = kobj;
    tcd->dev.devt =  MKDEV(resource_major[ISIL_CHIP],minor);
    tcd->dev.release = dumy_release;
    fix_id = get_chipid_from_epobj(&tcd->epobj);
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 30)
    dev_set_name(&tcd->dev, "%scontrol%02lu", tcd->name, fix_id);
#else
    snprintf(tcd->dev.bus_id,sizeof(tcd->dev.bus_id),"%scontrol%02lu",tcd->name,fix_id);
#endif
    ret = device_register(&tcd->dev);
    if(ret < 0){
        dbg("device register failed!\n");
        goto cleanup;
    }
    ret = bind_chip_driver(tcd);
    if(ret)
        goto bind_error;

    mutex_lock(&mutex_chipdev);
    list_add_tail(&tcd->list,&isil_chipdev);
    mutex_unlock(&mutex_chipdev);
    sema_init(&tcd->sem, 1);
    return ret;
bind_error:
    ISIL_DBG(ISIL_DBG_INFO, ": chip driver bind error!");
    device_unregister(&tcd->dev);
cleanup:
    if(kobj)
        kfree(kobj);
    if(tcd->cdev){
        if(flag == 0)
            cdev_del(tcd->cdev);
        kfree(tcd->cdev);
    }
    put_free_minor(tcd->minor,0);
    return ret;
}


/* need free all the sub resources such as tsd device...*/
/*******************************************************************************
* Function Name  : unregister_chip_device
* Description    : unregister one chip device's driver from OS
* Input          : struct isil_chip_device * tcd
* Output         : NULL
* Return         : NULL
*******************************************************************************/
void unregister_chip_device(struct isil_chip_device * tcd)
{
    struct kobject * dir_kobj = NULL;
    //int cnt;	

    dbg("passed tcd[0x%8p]\n",tcd);	
    if(!tcd){
        dbg("ungister with NULL pointer!\n");
        goto donothing;
    }
    if(find_chip_dev(tcd->minor) == NULL){
        dbg("bad ungister!\n");
        goto donothing;
    }

    dir_kobj = tcd->dev.kobj.parent; 
    if(!dir_kobj){
        dbg("this device is located wrong place!\n");
        goto donothing;
    }

    mutex_lock(&mutex_chipdev);
    list_del(&tcd->list);
    mutex_unlock(&mutex_chipdev);

    cdev_del(tcd->cdev);
    kfree(tcd->cdev);
    device_unregister(&tcd->dev);
    unbind_chip_driver(tcd);
    put_free_minor(tcd->minor,0);
    if(!tcd->clean_up){
        dbg("no clean_up function registed for tcd\n");
        isil_remove_dir(dir_kobj);
        dbg("after isil_remove_dir!\n");
        return;
    }
    tcd->clean_up(tcd);
    isil_remove_dir(dir_kobj);
    return;
donothing:
    return;	

}




/*******************************************************************************
* Function Name  : register_chip_driver
* Description    : register an chip device's driver
* Input          : struct isil_chip_driver * driv
* Output         : NULL
* Return         : 0:success; !0: failed
*******************************************************************************/
int register_chip_driver(struct isil_chip_driver * driv)
{
    int res = 0;

    driv->driver.bus = &isil_virtual_bus;
    INIT_LIST_HEAD(&driv->tcd_list);

    res = driver_register(&driv->driver);
    if (res)
        return res;
    mutex_lock(&mutex_chipdrv);
    list_add_tail(&driv->list,&isil_chipdrv);
    mutex_unlock(&mutex_chipdrv);

    return res;
}

/*
   int register_chip_driver_instance(struct isil_chip_driver * driv, struct isil_chip_device *tcd)
   {
   int res = 0;
   struct list_head *p;
   struct isil_chip_device * tcd = NULL;
   struct isil_dev_table_id * table_id = &driv->device_id;

   if(chip_device_match(&tcd->epobj,&table_id->tid->epobj)){
   tcd->driver = driv;
   if(driv->attach_chip)
   driv->attach_chip(driv,tcd);//driver init instance
   }

   return res;
   }
   */


/*******************************************************************************
* Function Name  : unregister_chip_driver
* Description    : unregister the chip driver, not the instance
* Input          : struct isil_chip_driver * driv
* Output         : NULL
* Return         : NULL
*******************************************************************************/
void unregister_chip_driver(struct isil_chip_driver * driv)
{
    driver_unregister(&driv->driver);
    mutex_lock(&mutex_chipdrv);	
    list_del(&driv->list);
    mutex_unlock(&mutex_chipdrv);
}


/*******************************************************************************
* Function Name  : check_valid
* Description    : check the argment is valid or not
* Input          : struct isil_chip_device* tcd,unsigned long type
* Output         : NULL
* Return         : 0: valid; -1: invalid
*******************************************************************************/
int check_valid(struct isil_chip_device* tcd,unsigned long type)
{
    if((type > MAX_SERVICE_NR) || tcd == NULL)
        return -1;

    return 0;
}


/*******************************************************************************
* Function Name  : check_chipdev_limit
* Description    : check the argment is valid or not
* Input          : struct isil_chip_device *tcd, struct cmd_arg *arg
* Output         : NULL
* Return         : 0: valid; <0: invalid; >0:calid
*******************************************************************************/
int check_chipdev_limit(struct isil_chip_device *tcd, struct cmd_arg *arg)
{
    struct list_head *p;
    struct cmd_list *cl;

    list_for_each(p,&cl_head){
        cl = container_of(p, struct cmd_list, node);
        if(cmd_equal(&cl->arg,arg) && cl->tcd == tcd){
           	/*return last gathered info to user*/
		memcpy((void *)&arg->dev_info2, (void *)&cl->di, sizeof(struct device_info));
        	return 1;
	}
    }
    return 0;

}

/*******************************************************************************
* Function Name  : get_dir
* Description    : create one dir in sys file system 
* Input          : struct isil_chip_device *tcd, unsigned long type, struct kobject **dir
* Output         : struct kobject *
* Return         : 0: means the dir is first created
                 : 1: means the dir is aready created brefore
                 : -1: failed
*******************************************************************************/
static int get_dir(struct isil_chip_device *tcd, unsigned long type, struct kobject **dir)
{
    char * name_prefix = NULL; 
    struct kobject *d = NULL;//tcd->dev.kobj.parent;	
    
    *dir = NULL;
    if(tcd->dirs[type]){
	*dir = tcd->dirs[type];
        return 1;
    }

    switch(type+1)
    {
        case ISIL_ENCODER:
            name_prefix = "isil_encoder";
            break;
        case ISIL_DECODER:
            name_prefix = "isil_decoder";
            break;
        case ISIL_VIDEO_IN:
            name_prefix = "isil_video_in";
            break;
        case ISIL_VIDEO_OUT:
            name_prefix = "isil_video_out";
            break;
        case ISIL_AUDIO_IN:
            name_prefix = "isil_audio_in";
            break;
        case ISIL_AUDIO_OUT:
            name_prefix = "isil_audio_out";
            break;
        default:
            name_prefix = "unknown";
            break;
    }

    /*arrive here, means no dir created , so we must create it here!*/
    d = isil_create_dir(name_prefix, tcd->dev.kobj.parent);
    if(d == NULL){
        dbg("create dir failed in %s\n", __FUNCTION__);
        return -1;
    }
//    kobject_uevent(d, KOBJ_ADD);
    *dir = d;
    return 0;
}





/*******************************************************************************
* Function Name  : lookup_tcdev_tabe_id
* Description    : find the tcdev_tabe_id by epobj_t
* Input          : struct epobj_t *ep
* Output         : NULL
* Return         : struct isil_dev_table_id *
*******************************************************************************/
static struct isil_dev_table_id * lookup_tcdev_tabe_id(struct epobj_t *ep)
{
    struct list_head *p = &isil_tc_dev;
    struct isil_dev_table_id	*dev;

    list_for_each(p, &isil_tc_dev){
        dev = list_entry(p, struct isil_dev_table_id, list);
        if (epobj_euqal(ep,&dev->tid->epobj))
            return dev;
    }
    return NULL;
}


/*******************************************************************************
* Function Name  : parse_arg
* Description    : parse the argment for creating driver instance
* Input          : struct isil_chip_device *tcd, struct cmd_arg *arg
* Output         : NULL
* Return         : 0:success; !0: failed
*******************************************************************************/
int parse_arg(struct isil_chip_device *tcd, struct cmd_arg *arg)
{
    chip_driver_t *cd;
    struct isil_logic_phy_table *tlp;
    int	ret = -1;
    enum  LOGIC_MAP_TABLE_TYPE  mt;

    if(!tcd || !arg || !tcd->ped){
        dbg("invalid parameters!\n");
        return -EINVAL;
    }


    cd = container_of(tcd->ped, chip_driver_t,opened_control_ed);
    tlp = &cd->logic_map_phy;
    if(!tlp->op || !tlp->op->find_phy_by_logic){
        ISIL_DBG(ISIL_DBG_INFO, ": logic map no op\n");
        return -EINVAL;
    }

    if(ISIL_ENCODER == arg->type){
        if(CODEC_VIDEO_H264 == arg->algorithm){
            mt = LOGIC_MAP_TABLE_BIND_H264E;
        }
        else if(STREAM_TYPE_AUDIO == arg->stream){
            mt = LOGIC_MAP_TABLE_BIND_AENC;
        }
        else if(CODEC_VIDEO_MJPG == arg->algorithm){
            mt = LOGIC_MAP_TABLE_BIND_JPEGE;
        }
        else{
            goto out;
        }
    }
    else if(ISIL_DECODER == arg->type){
        if(STREAM_TYPE_AUDIO == arg->stream){
            mt = LOGIC_MAP_TABLE_BIND_ADEC;
        }
        else if(CODEC_VIDEO_H264 == arg->algorithm){
            mt = LOGIC_MAP_TABLE_BIND_H264D;
        }
        else{
            goto out;	
        }
    }
    else{
        goto out;
    }
#if 0
    logic = arg->channel_idx;
    ret = tlp->op->find_phy_by_logic(tlp, mt,logic, &phy);
    if(ret){
        dbg("logic to phy find nothing!\n");
        goto out;
    }
    /*here we convert the logic channel id to phy id*/
    arg->channel_idx = phy;
#endif
    return 0;

out:
    return ret;	
}


/*******************************************************************************
* Function Name  : tc_driver_init
* Description    : create an instance of service driver for the crossponding tc_device
* Input          : struct isil_chip_device *tcd, struct isil_service_device *tsd,struct isil_service_driver *driv,
*                : struct cmd_arg *arg, unsigned long from
* Output         : NULL
* Return         : 0:success; !0: failed
*******************************************************************************/
int tc_driver_init(struct isil_chip_device *tcd, struct isil_service_device *tsd,struct isil_service_driver *driv,
					struct cmd_arg *arg, unsigned long from)
{
    int ret = 0;
    struct tsd_driver_operations *tsdr;
    struct isil_dev_table_id	*table_id;
    struct cmd_arg  tmp_arg = {0};

    if(!driv){
        ISIL_DBG(ISIL_DBG_INFO, ": --->in\n");
        return -EINVAL;
    }
    table_id = (struct isil_dev_table_id *)&driv->device_id;
    if(!table_id->ops){
        dbg("driver with no ops!\n");
        return -EINVAL;
    }
    tsdr = (struct tsd_driver_operations *)table_id->ops;
    if(!tsdr->init){
        dbg("driver with no init function!\n");
        return -EINVAL;
    }
    if(from){
    	memcpy((void *)&tmp_arg,(void *)arg, sizeof(struct cmd_arg));
    }else{
		if(copy_from_user((void *)&tmp_arg,(void __user *)arg, sizeof(struct cmd_arg)) != 0){
			ISIL_DBG(ISIL_DBG_INFO, ": copy_from_user failed\n");
		}
    }
    if(ISIL_ENCODER == arg->type || ISIL_DECODER == arg->type){
        if(parse_arg(tcd,&tmp_arg))
            return -EINVAL;
    }

    /*now the tmp_arg->channel_idx is the phy channel id, so the tc driver can use it to init the driver structures by phy id*/
    //printk("[[logic id = %lu]]\n",tmp_arg.channel_idx);

    //ret = tsdr->init(tcd,tsd,table_id->para,&tmp_arg);
    ret = tsdr->init(tcd,tsd,tcd_priv(tcd),&tmp_arg);
    if(ret){
        dbg("driver can not alloc resource!\n");
        return -ENOMEM;
    }
    epobj_key_update_phy_id((epkey_t *)&(tsd->epobj.key),tmp_arg.channel_idx);
    return ret;   
}




/*******************************************************************************
* Function Name  : setup_service_device
* Description    : core function for setup the service device 
* Input          : struct isil_service_device *tsd, struct isil_chip_device *tcd, struct cmd_arg *arg,
*                : struct kobject *dir,struct isil_dev_table_id *table_id, unsigned long from
* Output         : NULL
* Return         : 0:success; !0: failed
*******************************************************************************/
int setup_service_device(struct isil_service_device *tsd, struct isil_chip_device *tcd, struct cmd_arg *arg,
						struct kobject *dir,struct isil_dev_table_id *table_id, unsigned long from,
						struct device_info *di)
{
    int ret = 0,sma,minor;
    unsigned long isadd = 0;
    struct list_head *plist = NULL;
    int type = arg->type - 1;

    if(type < 0){
        ISIL_DBG(ISIL_DBG_INFO, ": BUG type error\n");
        return -1;
    }

    tcd->dirs[type] = dir;
    epobj_set(&tsd->epobj,&tcd->epobj);
    tsd->epobj.type = arg->type;
    tsd->epobj.func_id = (arg->stream << 16) | arg->algorithm;
    //tsd->epobj.key = make_service_key(tcd,arg);
    memset((char *)&tsd->epobj.key,0,sizeof(epkey_t));
    make_service_key((epkey_t *)&tsd->epobj.key,tcd,arg);
    //dbg_epobj_print(&tsd->epobj);
        if((ret = tc_driver_init(tcd,tsd,table_id->priv,arg, from))){
        ISIL_DBG(ISIL_DBG_INFO, ": service driver init failed!\n");
        return ret;
    }

    sma = resource_major[type + 1];
    if(type > 1){
	    snprintf(tsd->name, 20,"%s_%02lu",tc_dev_name[type+1],arg->channel_idx);
    }else{
        snprintf(tsd->name,sizeof(tsd->name),"chan%02lu-%s-%s",arg->channel_idx,stream[arg->stream],algorithm[arg->algorithm]);

	}
    tsd->cdev = cdev_alloc();
    if(tsd->cdev == NULL){
        dbg("cdev_alloc failed!\n");
        return -ENOMEM;
    }

    minor = get_free_minor(arg);
    if(minor < 0){
        dbg("service minor get failed!\n");
        kfree(tsd->cdev);
        return -EINVAL;
    }
    //dbg("the service device' major = %d; minor = %X\n",sma,minor);
    cdev_init(tsd->cdev,&tws_fops);
    tsd->cdev->owner = THIS_MODULE;
    ret = cdev_add(tsd->cdev,MKDEV(sma,minor),1);
    if(ret < 0){
        dbg("cdev_add failed!\n");
        goto error;
    }
    isadd = 1;

        INIT_LIST_HEAD(&tsd->list);
    tsd->minor = minor;
    tsd->ptcd = tcd;
    tsd->type = type;//type = arg->type -1
    tsd->dev.kobj.parent = dir;
    tsd->dev.class = NULL;
    tsd->dev.devt = MKDEV(sma,minor);
    tsd->dev.release = dumy_release;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 30)
    dev_set_name(&tsd->dev, "%s", tsd->name);
#else
    snprintf(tsd->dev.bus_id, 20,"%s",tsd->name);
#endif
    ret = device_register(&tsd->dev);
    if(ret<0){
        dbg("device_register failed in %s!\n",__FUNCTION__);
        goto error;
    }
#ifdef  STATIC_COUNTING
    memset((char *)&tsd->b_itval, 0, sizeof(struct tc_buf_interval));
    tsd->b_itval.u_time = 5;
#endif	
    plist = &tcd->tsd_list[type];// type = arg->type -1
    list_add_tail(&tsd->list,plist);

    di->major = sma;
    di->minor = minor;
    if(from == 2){
        memcpy(di->path,tsd->name,255);
	memcpy((void *)&arg->dev_info2,(void *)di,sizeof(struct device_info));
    }
    return 0;
error:
    if(tsd->cdev){
        if(isadd)
            cdev_del(tsd->cdev);
        kfree(tsd->cdev);
        put_free_minor(minor,arg->type);
    }	
    return ret;	

}


/*******************************************************************************
* Function Name  : tc_serdev_alloc
* Description    : allocate the memory for chip_device structure
* Input          : struct isil_chip_device * tcd, struct cmd_arg * arg, struct kobject * dir, unsigned long from
* Output         : NULL
* Return         : struct isil_service_device *
*******************************************************************************/
struct isil_service_device * tc_serdev_alloc(unsigned long priv_size)
{
    void * p;
    unsigned long size;
    struct isil_service_device *tsd;

    size = (sizeof(*tsd) + CHIPDEV_ALIGN_MASK) & ~CHIPDEV_ALIGN_MASK;
    size += priv_size + CHIPDEV_ALIGN_MASK;
    p = kzalloc(size, GFP_KERNEL);
    if(!p){
        ISIL_DBG(ISIL_DBG_INFO, ": no mem!\n");
        return NULL;
    }
    tsd = (struct isil_service_device *)(((long)p + CHIPDEV_ALIGN_MASK) & ~CHIPDEV_ALIGN_MASK);
    tsd->padded = (char *)tsd - (char *)p;
    if(priv_size)
        tsd->priv = tsd_priv(tsd);
    //dbg("####---%s p[0x%8p],sizeof(*tsd)[%d],tsd->priv[0x%8p]\n",__FUNCTION__,p,sizeof(struct isil_service_device),tsd->priv);
    return tsd;
} 

void tc_serdev_free(struct isil_service_device * tsd)
{
    kfree((char *)tsd - tsd->padded);
}


/*******************************************************************************
* Function Name  : do_create_service_dev
* Description    : create one service device,only setup essential staff in tsd for every devs
* Input          : struct isil_chip_device * tcd, struct cmd_arg * arg, struct kobject * dir, unsigned long from
* Output         : NULL
* Return         : 0:success; !0: failed
*******************************************************************************/
int do_create_service_dev(struct isil_chip_device * tcd, struct cmd_arg * arg, struct kobject * dir, unsigned long from)
{
    int ret = 0;
    struct epobj_t ep = {0};
    struct tsd_file_operations *tsd_ops = NULL;
    struct isil_service_device *tsd = NULL;
    struct isil_dev_table_id	*table_id;
    struct cmd_list  *cl = NULL;
    int type = arg->type - 1;
    struct device_info di;

    if(type < 0){
        ISIL_DBG(ISIL_DBG_INFO, ": BUG type error\n");
        return -1;
    }
    tcd->dirs[type] = dir;

    epobj_set(&ep,&tcd->epobj);
    ep.type = arg->type;
    ep.func_id = (arg->stream << 16) | arg->algorithm;
    //ep.key = make_service_key(tcd,arg);
    make_service_key((epkey_t *)&ep.key,tcd,arg);

    table_id = lookup_tcdev_tabe_id(&ep);
    if(!table_id){
        ISIL_DBG(ISIL_DBG_INFO, ": we didn't support this type device!\n");
        return -ENODEV;
    }
    tsd_ops = (struct tsd_file_operations *)table_id->ops;

    if(!tsd_ops->init){
        dbg("no init function for service device!\n");
        return -EINVAL;
    }

    /*allocate the customer service device's structure*/
    ret = tsd_ops->init(tcd,arg,&tsd);
    if(ret || !tsd){
        tsd_ops->destory(tsd);
        return -ENOMEM;
    }
    tsd->tsd_fops = table_id->ops;
    tsd->private_data = table_id->para;//tc_dev 's table_id->para
    /*setup tsd, and call crosponding service driver's init*/
    memset(&di, 0, sizeof(struct device_info));
    ret = setup_service_device(tsd,tcd,arg,dir,table_id,from,&di);
    if(ret){
        tsd_ops->destory(tsd);
        return ret;
    }

    /*if the service device create success, we can add it to cl_head for later(release) use*/
    cl = kzalloc(sizeof(*cl),GFP_KERNEL);
    if(!cl){
        dbg("cmd_list alloc failed!\n");
        ret = -ENOMEM;
        goto d_out;
        /*fall through???*/      
    }

    INIT_LIST_HEAD(&cl->node);
    cl->tcd = tcd;
    memcpy((char *)&cl->arg,(char *)arg, sizeof(struct cmd_arg));
    memcpy((char *)&cl->di, (char *)&di, sizeof(struct device_info));
    list_add_tail(&cl->node,&cl_head);
    //printk("cmd_type[%d],cmd_cix[%lu],cmd_algo[%d],cmd_stream[%d]\n",cl->arg.type,cl->arg.channel_idx,cl->arg.algorithm,cl->arg.stream);
    return 0;
d_out:
    release_service_device(tcd, (void *)arg);
    return ret;
}



/*******************************************************************************
* Function Name  : create_service_device
* Description    : create one service device
* Input          : struct isil_chip_device * tcd, unsigned long data, unsigned long from
* Output         : NULL
* Return         : 0:success; !0: failed
*******************************************************************************/
int create_service_device(struct isil_chip_device * tcd, void *data, unsigned long from)
{
	int ret = 0,tmp;
	struct kobject *dir = NULL;
	struct cmd_arg *arg = (struct cmd_arg *)data;
	int it;

	it = arg->type -1;
	if(it < 0){
		ISIL_DBG(ISIL_DBG_INFO, ": BUG error type\n");
		return -EINVAL;
	}
	if(check_valid(tcd,arg->type)){
		dbg("invalid parameters\n");
		return -EINVAL;
	}
	
	if(check_chipdev_limit(tcd, arg)){
		ISIL_DBG(ISIL_DBG_INFO, ": have been created before!\n");
		return 0;
	}
	tmp = get_dir(tcd, it, &dir);
	if(dir == NULL){
		dbg("get dir failed\n");
		return -EEXIST; //this return value is not correct, doesn't means the dir is already exits.here just return an value < 0
	}
	ret = do_create_service_dev(tcd, arg,dir, from);
	if(ret && tmp == 0) {
		isil_remove_dir(dir);
	}
	return ret;
}



/*******************************************************************************
* Function Name  : tc_driver_instant_remove
* Description    : remove one service driver instance
* Input          : struct isil_chip_device *tcd, struct isil_service_device *tsd,struct isil_service_driver *driv, struct cmd_arg *arg
* Output         : NULL
* Return         : NULL
*******************************************************************************/
void tc_driver_instant_remove(struct isil_chip_device *tcd, struct isil_service_device *tsd,struct isil_service_driver *driv, struct cmd_arg *arg)
{
    struct tsd_driver_operations *tsdr;
    struct isil_dev_table_id	*driv_table_id;

    if(!driv){
        ISIL_DBG(ISIL_DBG_INFO, ": NULL pointer\n");
        return;
    }
    driv_table_id = (struct isil_dev_table_id *)&driv->device_id;
    if(!driv_table_id->ops){
        dbg("driver with no ops!\n");
        return;
    }
    tsdr = (struct tsd_driver_operations *)driv_table_id->ops;
    if(!tsdr->remove){
        dbg("driver with no remove function!\n");
        return;
    }
    tsdr->remove(tcd,tsd,tcd_priv(tcd),arg);
}



/*******************************************************************************
* Function Name  : find_tc_serdriv_from_tcd
* Description    : find the dev_table_id from one chip_device by given argment
* Input          : struct isil_service_device *tsd,void *data
* Output         : NULL
* Return         : the desired struct isil_dev_table_id	*
*******************************************************************************/
struct isil_dev_table_id	*find_tc_serdriv_from_tcd(struct isil_chip_device *tcd,struct cmd_arg *arg)
{
        struct epobj_t ep = {0};

        epobj_set(&ep,&tcd->epobj);
	ep.type = arg->type;
	ep.func_id = (arg->stream << 16) | arg->algorithm;
	//ep.key = make_service_key(tcd,arg);
        make_service_key((epkey_t *)&ep.key,tcd,arg);

        //dbg_epobj_print(&ep);
        return lookup_tcdev_tabe_id(&ep);
        
}


/*******************************************************************************
* Function Name  : do_release_service_device
* Description    : the core function performs service device releasing
* Input          : struct isil_service_device *tsd,void *data
* Output         : NULL
* Return         : NULL
*******************************************************************************/
void do_release_service_device(struct isil_service_device *tsd,void *data)
{
    struct kobject *k;
    struct isil_chip_device *t;
    struct list_head *p,*tmp;
    struct cmd_list *cl;
    struct cmd_arg *arg = (struct cmd_arg *)data;
    struct isil_dev_table_id	*table_id;

    k = tsd->dev.kobj.parent;
    cdev_del(tsd->cdev);
    kfree(tsd->cdev);
    device_unregister(&tsd->dev);
    //printk("tsd_type +1 = %d\n",tsd->type + 1);
    put_free_minor(tsd->minor,tsd->type + 1);
    t = tsd->ptcd;
    if(!t)
        dbg("BUG! in %s tsd has inside none tcd!\n",__FUNCTION__);
    if(data){
        if(list_empty(&t->tsd_list[tsd->type])){
            dbg("remove service's parent!\n");              
            isil_remove_dir(k);
            t->dirs[tsd->type] = NULL; 
        }
        list_for_each_safe(p, tmp, &cl_head){
            cl = container_of(p, struct cmd_list, node);
            if(cmd_equal(&cl->arg, arg) && cl->tcd == t){
                list_del(p);
                kfree(cl);
                break;
            }
        }
    }

    table_id = find_tc_serdriv_from_tcd(t,arg);
    if(table_id){
        dbg("remove driver instance!\n");
        tc_driver_instant_remove(t,tsd,table_id->priv,arg);
    }else{
        dbg("can not find driver in tc_dev list\n");
    }

    /*perform future release*/
    if(!tsd->clean_up){
        dbg("you guys have no clean_up function registed\n");
        goto out;
    }
    tsd->clean_up(tsd);
 out:
    ISIL_DBG(ISIL_DBG_INFO, ": after perforem release service -device tsd[0x%8p]\n\n\n\n\n",tsd);
    return;
}




/*******************************************************************************
* Function Name  : release_service_device
* Description    : the wapper function performs service device releasing
* Input          : struct isil_chip_device *tcd, void *data
* Output         : NULL
* Return         : NULL
*******************************************************************************/
void  release_service_device(struct isil_chip_device *tcd, void *data)
{
    struct cmd_arg *arg = (struct cmd_arg *)data;
    epkey_t new_key = {0};
    struct isil_service_device *tmp;
    struct list_head *p,*head;
    int type = arg->type - 1;

    tc_trace;
    if(type < 0){
        ISIL_DBG(ISIL_DBG_INFO, ": BUG: error type\n");
        return;
    }
    mutex_lock(&mutex_chipdev);	
    make_service_key((epkey_t *)&new_key,tcd,arg);
    head = &tcd->tsd_list[type];
    list_for_each(p,head){
        tmp = container_of(p,struct isil_service_device,list);
		if(tmp && epkey_equal_without_phyid(&tmp->epobj.key,&new_key)){
			goto find;
        }
    }
    ISIL_DBG(ISIL_DBG_INFO, ": @@@@@@@@@@---->invalid parameter! \n");
    ISIL_DBG(ISIL_DBG_INFO, ": @@@@@@@@@@type[%d],cix[%lu],algo[%d],stream[%d]\n",arg->type,arg->channel_idx,arg->algorithm,arg->stream);
    mutex_unlock(&mutex_chipdev);	
    return;
find:
    list_del(&tmp->list);
    mutex_unlock(&mutex_chipdev);

    do_release_service_device(tmp,data);

    return ;
}


/*******************************************************************************
* Function Name  : release_all_service_device
* Description    : release all service devices in tc_core
* Input          : NULL
* Output         : NULL
* Return         : NULL
*******************************************************************************/
void release_all_service_device(void)
{
    struct list_head  *p,*q,*tmp,*tmt;
    //struct list_head  remove_list;
    struct isil_chip_device * tcd = NULL;
    struct isil_service_device *tsd = NULL;
    int idx = 0;
    struct kobject *k;
    struct cmd_list *cl;

    LIST_HEAD(rm_list);
    INIT_LIST_HEAD(&rm_list);

    mutex_lock(&mutex_chipdev);
    list_for_each(p,&isil_chipdev){
        tcd = container_of(p,struct isil_chip_device, list);
        for(idx=0; idx < MAX_SERVICE_NR-1; idx++){
            q = &tcd->tsd_list[idx];
            if(list_empty(q))	
                continue;
            list_for_each_safe(tmp,tmt,q){
                list_move_tail(tmp,&rm_list);
            }
        }
    }
    mutex_unlock(&mutex_chipdev);


    list_for_each_safe(p, tmp, &rm_list){
        //printk("----->in %s\n",__FUNCTION__);
        list_del(p);  
        tsd = list_entry(p, struct isil_service_device, list);
        do_release_service_device(tsd,NULL);//use NULL in internal of this function will never touch the tcd->dirs, 
    } 

    /*arrive here all service device have been removed, we need do more things
     *1: clean cmd_list 
     *2: update all the service device's info in tcd
     */
    if(!list_empty(&rm_list))
        ISIL_DBG(ISIL_DBG_INFO, ": BUG have service_dev lefted!\n");

    list_for_each_safe(p,tmp,&cl_head){
        list_del(p);
        cl = container_of(p, struct cmd_list, node);
        kfree(cl);
    }

    mutex_lock(&mutex_chipdev);
    list_for_each(p,&isil_chipdev){
        tcd = container_of(p,struct isil_chip_device, list);
        for(idx=0; idx < MAX_SERVICE_NR-1; idx++){
            //q = &tcd->tsd_list[idx];
            k = tcd->dirs[idx];
            //if(list_empty(q)){	
            if(k){
                isil_remove_dir(k);

            }
            //}
            tcd->dirs[idx] = NULL;
        }
    }
    mutex_unlock(&mutex_chipdev);
    return;
}

/*******************************************************************************
* Function Name  : release_all_service_device_on_chip
* Description    : release all service device on one chip_device
* Input          : struct isil_chip_device *tcd
* Output         : NULL
* Return         : NULL
*******************************************************************************/
void release_all_service_device_on_chip(struct isil_chip_device *tcd)
{
    struct list_head  *p,*q,*tmp,*tmt;
    struct isil_service_device *tsd = NULL;
    int idx = 0;
    struct kobject *k;
    struct cmd_list *cl;

    LIST_HEAD(rm_list);
    INIT_LIST_HEAD(&rm_list);

    if(!tcd)
        return;


    mutex_lock(&mutex_chipdev);
    for(idx=0; idx < MAX_SERVICE_NR-1; idx++){
        q = &tcd->tsd_list[idx];
        if(list_empty(q))	
            continue;
        list_for_each_safe(tmp,tmt,q){
            list_move_tail(tmp,&rm_list);
        }
    }
    mutex_unlock(&mutex_chipdev);


    list_for_each_safe(p, tmp, &rm_list){
        list_del(p);  
        tsd = list_entry(p, struct isil_service_device, list);
        do_release_service_device(tsd,NULL);//use NULL in internal of this function will never touch the tcd->dirs, 
    } 

    /*arrive here all service device have been removed, we need do more things
     *1: clean cmd_list 
     *2: update all the service device's info in tcd
     */
    if(!list_empty(&rm_list))
        ISIL_DBG(ISIL_DBG_INFO, ": BUG have service_dev lefted!\n");

    list_for_each_safe(p,tmp,&cl_head){
        cl = container_of(p, struct cmd_list,node);
        if(cl->tcd == tcd){
            list_del(p);
            kfree(cl);
        }
    }

    mutex_lock(&mutex_chipdev);
    for(idx=0; idx < MAX_SERVICE_NR-1; idx++){
        //q = &tcd->tsd_list[idx];
        k = tcd->dirs[idx];
        //if(list_empty(q)){	
        if(k){
            isil_remove_dir(k);

        }
        //}
        tcd->dirs[idx] = NULL;
    }
    mutex_unlock(&mutex_chipdev);
    return;
}





/*******************************************************************************
* Function Name  : release_all_service_device_ext
* Description    : release all service devices in tc_core
* Input          : NULL
* Output         : NULL
* Return         : NULL
*******************************************************************************/
void release_all_service_device_ext(void)
{
    struct list_head  *p,*q,*tmp;
    struct isil_chip_device * tcd = NULL;
    struct cmd_list *cl;

    mutex_lock(&mutex_chipdev);
    list_for_each(p,&isil_chipdev){
        tcd = container_of(p,struct isil_chip_device, list);
        list_for_each_safe(q,tmp,&cl_head){
            cl = container_of(q,struct cmd_list,node);
            if(cl->tcd == tcd){
                mutex_unlock(&mutex_chipdev);       
                release_service_device(cl->tcd,(void *)&cl->arg);
                mutex_lock(&mutex_chipdev);
            }
        }
    }
    mutex_unlock(&mutex_chipdev);

    /*since we call release_service_device with cl->arg as the second parameter, so we need do nothing here */
    return;
}


/*******************************************************************************
* Function Name  : release_all_service_device_ext_on_chip
* Description    : release all service device on one chip_device
* Input          : struct isil_chip_device *tcd
* Output         : NULL
* Return         : NULL
*******************************************************************************/
void release_all_service_device_ext_on_chip(struct isil_chip_device *tcd)
{
    struct list_head  *q,*tmp;
    struct cmd_list *cl;

    if(!tcd)
        return;
    mutex_lock(&mutex_chipdev);
    list_for_each_safe(q,tmp,&cl_head){
        cl = container_of(q,struct cmd_list,node);
        if(cl->tcd == tcd){
            mutex_unlock(&mutex_chipdev);       
            release_service_device(cl->tcd,(void *)&cl->arg);
            mutex_lock(&mutex_chipdev);
        }
    }
    mutex_unlock(&mutex_chipdev);

    /*since we call release_service_device with cl->arg as the second parameter, so we need do nothing here */
    return;
}



/*******************************************************************************
* Function Name  : register_tc_device
* Description    : register one tc_device into tc_core
* Input          : struct isil_dev_table_id *new
* Output         : NULL
* Return         : 0:success; !0: failed
*******************************************************************************/
int register_tc_device(struct isil_dev_table_id *new)
{
    int ret = 0;
    struct list_head *p = &isil_tc_dev;
    struct isil_dev_table_id	*tb,*prev;


    mutex_lock(&mutex_tc_dev);
    if(p->next == &isil_tc_dev && p->prev == &isil_tc_dev){
        list_add_tail(&new->list,&isil_tc_dev);
        ret = 0;
        goto out;
    }

    p = p->next;
    if(p == &isil_tc_dev){
        list_add_tail(&new->list,&isil_tc_dev);
        ret = 0;
        goto out;
    }

    for(;p != &isil_tc_dev; prev = tb, p = p->next)
    {
        tb = container_of(p,struct isil_dev_table_id,list);
        if(tb->tid->epobj.vendor_id != new->tid->epobj.vendor_id || tb->tid->epobj.bus_id != new->tid->epobj.bus_id 
                || tb->tid->epobj.chip_id != new->tid->epobj.chip_id || tb->tid->epobj.func_id != new->tid->epobj.func_id)
            continue;
        if(tb->tid->version != new->tid->version){
            ISIL_DBG(ISIL_DBG_INFO, ": already exist!\n");
            ret = -EEXIST;
            goto out;
        }
	/*add by version order, the bigger one goes first */
        if(new->tid->version > tb->tid->version){
            list_add(&new->list,&prev->list); 
            ret = 0;
            goto out;
        }
    }
    dbg_epobj_print(&new->tid->epobj);
    list_add(&new->list,&prev->list);
out:
    mutex_unlock(&mutex_tc_dev);
    return ret;	

}



/*******************************************************************************
* Function Name  : unregister_all_tc_device
* Description    : unregister all the tc_device registered in tc_core
* Input          : NULL
* Output         : NULL
* Return         : NULL
*******************************************************************************/
void unregister_all_tc_device(void)
{
    struct list_head *_p,*tmp;
    struct isil_dev_table_id *dev;

    LIST_HEAD(remove_list);
    INIT_LIST_HEAD(&remove_list);

    mutex_lock(&mutex_tc_dev);
    while(!list_empty(&isil_tc_dev)){
        list_for_each_safe(_p, tmp, &isil_tc_dev){
            //dev = list_entry(_p, struct isil_dev_table_id, list);
            list_move_tail(_p,&remove_list);
        }
    }
    mutex_unlock(&mutex_tc_dev);


    list_for_each_safe(_p, tmp, &remove_list){
        dbg("now deleting...\n");
        list_del(_p);  
        dev = list_entry(_p, struct isil_dev_table_id, list);
        kfree(dev);               
    } 
    if(!list_empty(&remove_list))
        ISIL_DBG(ISIL_DBG_INFO, ": BUG: have tc_dev lefted\n");    
    return;
}




/*******************************************************************************
* Function Name  : unregister_tc_device
* Description    : unregister one tc_device from tc_core
* Input          : struct isil_dev_table_id *tb
* Output         : NULL
* Return         : NULL
*******************************************************************************/
void unregister_tc_device(struct isil_dev_table_id *tb)
{
    mutex_lock(&mutex_tc_dev);
    list_del(&tb->list);
    mutex_unlock(&mutex_tc_dev);
    kfree(tb);
}




/*******************************************************************************
* Function Name  : tc_driver_match
* Description    : match the driver and the device by dev_table_id
* Input          : struct isil_dev_table_id *driv,struct isil_dev_table_id *dev
* Output         : NULL
* Return         : 0: not match; 1: match success
*******************************************************************************/
int tc_driver_match(struct isil_dev_table_id *driv,struct isil_dev_table_id *dev)
{
    if(epobj_euqal(&driv->tid->epobj,&dev->tid->epobj))                
        return 1;
    else
        return 0;
}



/*******************************************************************************
* Function Name  : register_tc_driver
* Description    : register one tc driver into tc_core
* Input          : struct isil_service_driver *driv
* Output         : NULL
* Return         : 0: success; !0: failed
*******************************************************************************/
int register_tc_driver(struct isil_service_driver *driv)
{
    int ret= 0;
    struct list_head *_p;
    struct isil_dev_table_id *dev;

    dev = &driv->device_id;
    dbg_epobj_print(&dev->tid->epobj);
    driv->driver.bus = &isil_virtual_bus;
    ret = driver_register(&driv->driver);
    if(ret){
        if(-EEXIST != ret)
            return ret;
        else
            return 0;
    }
    mutex_lock(&mutex_tc_drv);
    list_add_tail(&driv->list,&isil_tc_drv);
    mutex_unlock(&mutex_tc_drv);

    list_for_each(_p, &isil_tc_dev){
        dev = list_entry(_p, struct isil_dev_table_id, list);
        if (tc_driver_match(&driv->device_id,dev)){
            //printk("attached!\n");
            dev->priv = driv;   //tc_dev->table_id->priv = isil_service_driver
            break;
        }
    }
    return ret;	
}




/*******************************************************************************
* Function Name  : unregister_tc_driver
* Description    : unregister one tc driver form tc_core
* Input          : struct isil_service_driver *driv
* Output         : NULL
* Return         : NULL
*******************************************************************************/
void unregister_tc_driver(struct isil_service_driver *driv)
{
    struct list_head *_p;
    struct isil_dev_table_id *dev;

    list_for_each(_p, &isil_tc_dev){
        dev = list_entry(_p, struct isil_dev_table_id, list);
        if (tc_driver_match(&driv->device_id,dev)){
            dev->priv = NULL;
            break;
        }
    }        
    driver_unregister(&driv->driver);
    mutex_lock(&mutex_tc_drv);
    list_del(&driv->list);
    mutex_unlock(&mutex_tc_drv);
    ISIL_DBG(ISIL_DBG_INFO, "\n");
}



/*export the helper functions to other*/
EXPORT_SYMBOL(register_chip_device);
EXPORT_SYMBOL(unregister_chip_device);

EXPORT_SYMBOL(register_chip_driver);
EXPORT_SYMBOL(unregister_chip_driver);

EXPORT_SYMBOL(create_service_device);
EXPORT_SYMBOL(release_service_device);

EXPORT_SYMBOL(register_tc_device);
EXPORT_SYMBOL(unregister_tc_device);

EXPORT_SYMBOL(register_tc_driver);
EXPORT_SYMBOL(unregister_tc_driver);

EXPORT_SYMBOL(find_chip_dev);
EXPORT_SYMBOL(find_service_dev);

EXPORT_SYMBOL(tc_chipdev_alloc);
EXPORT_SYMBOL(tc_chipdev_free);






/*******************************************************************************
* Function Name  : request_devno_resource
* Description    : request the device major and minior resource from OS
* Input          : NULL
* Output         : NULL
* Return         : 0:success; !0:failed
*******************************************************************************/
static int request_devno_resource()
{
    int ret = -1,i,j = 0;

    ret = init_nfm_bitmap();
    if(-1 == ret)
        return -1;
    for(i = ISIL_MAJOR_SEARCH_BASE; i < ISIL_MAJOR_MAX && j < MAX_SERVICE_NR; i++){
        ret = register_chrdev_region(MKDEV(i,0), MAX_MINOR_NR, tc_dev_name[j]);
        if (ret >= 0)
        	resource_major[j++] = i;
    }
    if(j != MAX_SERVICE_NR)
    	goto fail;
    return 0;
fail:
    for(i = 0; i < j; i++){
        unregister_chrdev_region(MKDEV(resource_major[i],0), MAX_MINOR_NR);
        free_nfm_bitmap();
    }
    return -1;
}


/*******************************************************************************
* Function Name  : free_devno_resource
* Description    : free the device major and minior to OS
* Input          : NULL
* Output         : NULL
* Return         : NULL
*******************************************************************************/
static void free_devno_resource()
{
    unsigned long i;

    free_nfm_bitmap();
    for(i = 0; i < MAX_SERVICE_NR; i++)
        unregister_chrdev_region(MKDEV(resource_major[i],0), MAX_MINOR_NR);
}


/*******************************************************************************
* Function Name  : tc_header_buffer_init
* Description    : setup the header buffer
* Input          : struct tc_hb_pool *pool,struct tc_header_buf *thb, int index
* Output         : NULL
* Return         : 0:success; !0: failed
*******************************************************************************/
int tc_header_buffer_init(struct tc_hb_pool *pool,struct tc_header_buf *thb, int index)
{
    tcb_node_t	*node;

    if(!pool || !thb){
        ISIL_DBG(ISIL_DBG_INFO, ": invalid para\n");
        return -EINVAL;
    }
    atomic_set(&thb->ref, 0);
    node = &thb->node;
    node->op = &tcb_node_op;
    node->op->init(node);
    node->op->set_priv(node, thb);
    spin_lock_init(&thb->lock);
    thb->id = index;
    thb->unit_size= PAGE_SIZE;
    thb->data = (unsigned long *)__get_free_page(GFP_KERNEL);
    if(thb->data == NULL) {
        ISIL_DBG(ISIL_DBG_ERR, "no memory\n");
        return -EINVAL;
    }
    if(thb->data != NULL) {
        memset((char *)thb->data,0,thb->unit_size);
    }
    return 0;
}



/*******************************************************************************
* Function Name  : tc_header_buffer_release
* Description    : release the header buffer
* Input          : struct tc_hb_pool *pool,struct tc_header_buf *thb
* Output         : NULL
* Return         : NULL
*******************************************************************************/
void tc_header_buffer_release(struct tc_hb_pool *pool,struct tc_header_buf *thb)
{
    struct tc_header_buf * tb;
    tcb_node_t * node;
    unsigned long flags;	

    if(thb){
        tb = thb;
        if(!pool)
            pool = tb->pool;

        spin_lock_irqsave(&tb->lock, flags);
        if(atomic_read(&tb->ref) > 1){
            atomic_dec(&tb->ref);
        }
        else
        {
            //if(tb->op && tb->op->reset())
            //	tb->op->reset();
            node = &tb->node;
            if(node->op){
                //dbg("pool =%p and ->release =%p\n",&pool->np,node->op->release);
                node->op->release(node,&pool->np);
                //dbg("after!\n");
            }
        }
        //*thb = NULL;
        spin_unlock_irqrestore(&tb->lock, flags);
    }
}

struct tc_hb_operation    tc_hb_op =
{
    .init = tc_header_buffer_init,
    .release = tc_header_buffer_release,
};


/*******************************************************************************
* Function Name  : tc_create_header_buffer_pool
* Description    : allocate the header buffer pool
* Input          : unsigned long len
* Output         : NULL
* Return         : 0 success; !0: failed
*******************************************************************************/
int tc_create_header_buffer_pool(unsigned long len)
{
    int ret = 0,idx, nr;
    tcb_node_pool_t	*np;
    struct tc_header_buf *thb;

    thb_pool->order = get_order(len);
    nr = len/TC_HEADER_BUFFER_LEN;
    np = &thb_pool->np;
    np->op = &tcb_node_pool_op;
    thb_pool->bufs = kzalloc(sizeof(struct tc_header_buf) * nr, GFP_KERNEL);
    if(!thb_pool->bufs){
        ret = -ENOMEM;
        return ret;
    }
    np->op->init(np, nr);
    thb = thb_pool->bufs;
    for(idx = 0; idx < nr; idx++,thb++, thb_pool->nr++){
        thb->op = &tc_hb_op;
        thb->pool = thb_pool;
        ret = thb->op->init(thb_pool,thb,idx);
        if(ret < 0){
            np->op->init(np, 0);
            ISIL_DBG(ISIL_DBG_ERR, "init thb failed %d\n", ret);
            break;
        }
        thb->op->release(thb_pool,thb);
    }
    if(ret < 0){
        for(idx = 0; idx < thb_pool->nr; idx++){
            thb = &thb_pool->bufs[idx];
            if(thb->data != NULL){
                free_page((unsigned long)thb->data);
                thb->data = NULL;
            }
        }
        kfree(thb_pool->bufs);
        thb_pool->bufs = NULL;
        return ret;
    }
    return ret;
}



/*******************************************************************************
* Function Name  : tc_destory_header_buffer_pool
* Description    : release the header buffer pool to OS
* Input          : struct tc_hb_pool *tp
* Output         : NULL
* Return         : NULL
*******************************************************************************/
void tc_destory_header_buffer_pool(struct tc_hb_pool *tp)
{
    tcb_node_pool_t *pn;

    if(!tp)
        return;

    pn = &tp->np;
    if(pn->op){
        pn->op->release(pn);
    }
    if(tp->bufs){
        struct tc_header_buf *thb;
        int i;
        for(i=0; i<tp->nr; i++) {
            thb = &tp->bufs[i];
            if(thb->data != NULL) {
                free_page((unsigned long)thb->data);
                thb->data = NULL;
            }
        }
        kfree(tp->bufs);
    }
    tp->nr = 0;
    tp->bufs = NULL;
    tp->op = NULL;
}

/*******************************************************************************
* Function Name  : tc_get_header_buf
* Description    : request one header buffer from the pool,wait untill has buffer
* Input          : struct tc_hb_pool *tp
* Output         : struct tc_header_buf **thb
* Return         : NULL
*******************************************************************************/
void tc_get_header_buf(struct tc_hb_pool *tp, struct tc_header_buf **thb)
{
	tcb_node_pool_t *tnp;
	tcb_node_t *tn;
	unsigned long fg;

	if(!tp || !thb){
		ISIL_DBG(ISIL_DBG_INFO, ": invalid parameter\n");
		return;
	}
	spin_lock_irqsave(&tp->lock, fg);
	tnp = &tp->np;
	*thb = NULL;
	if(!tnp->op){
		ISIL_DBG(ISIL_DBG_INFO, ": pool node has no op!\n");
		goto out;
	}
	tnp->op->get(tnp,&tn);
	if(tn){
		*thb = get_thb_from_node(tn);
		atomic_inc(&((*thb)->ref));
	}
	else
	{
		*thb = NULL;
	}
out:        
	spin_unlock_irqrestore(&tp->lock, fg);
}


/*******************************************************************************
* Function Name  : tc_try_get_header_buf
* Description    : request one header buffer from the pool,not wait if has no buffer
* Input          : struct tc_hb_pool *tp
* Output         : struct tc_header_buf **thb
* Return         : NULL
*******************************************************************************/
void tc_try_get_header_buf(struct tc_hb_pool *tp, struct tc_header_buf **thb)
{
	tcb_node_pool_t *tnp;
	tcb_node_t *tn;
	unsigned long fg;

	if(!tp || !thb){
		ISIL_DBG(ISIL_DBG_INFO, ": invalid parameter\n");
		return;
	}

	spin_lock_irqsave(&tp->lock, fg);  
	tnp = &tp->np;
	*thb = NULL;
	if(!tnp->op){
		ISIL_DBG(ISIL_DBG_INFO, ": pool node has no op!\n");
		goto out;
	}
	tnp->op->try_get(tnp,&tn);
	if(tn){
		*thb = get_thb_from_node(tn);
		atomic_inc(&((*thb)->ref));
	}
	else
	{
		*thb = NULL;
	}
out:
	spin_unlock_irqrestore(&tp->lock, fg);

}


/*******************************************************************************
* Function Name  : tc_put_header_buf
* Description    : return header buffer to the pool
* Input          : struct tc_hb_pool *tp, struct tc_header_buf *thb
* Output         : NULL
* Return         : NULL
*******************************************************************************/
void tc_put_header_buf(struct tc_hb_pool *tp, struct tc_header_buf *thb)
{
	tcb_node_pool_t *tnp;
	unsigned long fg;

	if(!tp || !thb){
		ISIL_DBG(ISIL_DBG_INFO, ": invalid parameter\n");
		return;
	}
	spin_lock_irqsave(&tp->lock, fg);         
	tnp = &tp->np;
	if(!tnp->op){
		ISIL_DBG(ISIL_DBG_INFO, ": pool node has no op!\n");
		goto out;
	} 
	tnp->op->put(tnp,&thb->node);
out:
	spin_unlock_irqrestore(&tp->lock, fg);

}

/*******************************************************************************
* Function Name  : tc_get_hb_pool_entry_number
* Description    : get the number of header buffer
* Input          : struct tc_hb_pool
* Output         : NULL
* Return         : the number of header buffer
*******************************************************************************/
int tc_get_hb_pool_entry_number(struct tc_hb_pool *tp)
{
	tcb_node_pool_t *tnp;

    if(!tp){
        ISIL_DBG(ISIL_DBG_INFO, ": invalid parameter in %s!\n",__FUNCTION__);
        return EINVAL;
    }
    tnp = &tp->np;
    if(!tnp->op){
        ISIL_DBG(ISIL_DBG_INFO, ": pool node has no op!\n");
        return EINVAL;
    }
    return tnp->op->get_curr_pool_entry_number(tnp);        
}

/*head buffer pool operation*/
struct tc_hb_pool_operation    thb_op = 
{
    .create                         = tc_create_header_buffer_pool,
    .destory                        = tc_destory_header_buffer_pool,
    .get_header_buf                 = tc_get_header_buf,
    .try_get_header_buf             = tc_try_get_header_buf,
    .put_header_buf                 = tc_put_header_buf,
    .get_hb_pool_entry_number       = tc_get_hb_pool_entry_number,
};

/*******************************************************************************
* Function Name  : setup_thb_pool
* Description    : setup the header buffer pool
* Input          : NULL
* Output         : NULL
* Return         : 0: success; !0: failed
*******************************************************************************/
int setup_thb_pool(void)
{
    int ret = 0;

    thb_pool = kzalloc(sizeof(*thb_pool),GFP_KERNEL);
    if(!thb_pool)
        return -ENOMEM;
    spin_lock_init(&thb_pool->lock);
    thb_pool->op = &thb_op;
    ret = thb_pool->op->create(TC_HEADER_BUFFER_POOL_LEN);
    return ret;
}

/*******************************************************************************
* Function Name  : cleanup_thb_pool
* Description    : remove and release the header buffer pool
* Input          : struct tc_hb_pool
* Output         : NULL
* Return         : NULL
*******************************************************************************/
void cleanup_thb_pool(struct tc_hb_pool *tp)
{
    if(tp == NULL){
        ISIL_DBG(ISIL_DBG_INFO, ": invalid parameter! in %s\n",__FUNCTION__);
        return;
    }

    if(tp->op != NULL)
        tp->op->destory(tp);
    else 
        tc_destory_header_buffer_pool(tp);

    kfree(tp);
}


/*******************************************************************************
* Function Name  : isil_core_init
* Description    : setup the device layer core structures
* Input          : NULL
* Output         : NULL
* Return         : 0: success; !0: failed
*******************************************************************************/
int isil_core_init(void)
{
    int ret = -1;

    INIT_LIST_HEAD(&isil_chipdev);
    INIT_LIST_HEAD(&isil_chipdrv);
    INIT_LIST_HEAD(&isil_tc_dev);
    INIT_LIST_HEAD(&isil_tc_drv);
    ret = bus_register(&isil_virtual_bus);
    if(ret){
        dbg("isil_virtual bus register failed!\n");
        return ret;
    }

    if(request_devno_resource())
        goto err_devno;
    if(setup_thb_pool())
        goto err_hb_pool;	

    kroot = isil_create_dir("isil",NULL);
    if(kroot == NULL) {
		dbg("Err: failed in root dir isil!\n");
        goto err_isil_dir;
    }
    return 0;

err_isil_dir:
    isil_remove_dir(kroot);
err_hb_pool:
    cleanup_thb_pool(thb_pool);
err_devno:
    free_devno_resource();
    bus_unregister(&isil_virtual_bus);	
    return ret;
}


/*******************************************************************************
* Function Name  : isil_core_exit
* Description    : removeand release the device layer core structures
* Input          : NULL
* Output         : NULL
* Return         : NULL
*******************************************************************************/
/*static*/ void /*__exit*/ isil_core_exit(void)
{
    isil_remove_dir(kroot);
    cleanup_thb_pool(thb_pool);
    free_devno_resource();
    bus_unregister(&isil_virtual_bus);
}

//module_init(isil_core_init);
//module_exit(isil_core_exit);



