/******************** (C) COPYRIGHT 2010 Intersil ********************
* File Name          : tc_chip_ai.c
* Author             : zhang zhao
* Revision           : 1.2.1.1289
* Date               : 2012-04-05
* Description        : the tc_dev's staff for audio input
* Change log         : none
********************************************************************************
* THE PRESENT SOFTWARE IS DISTRIBUTED IN THE HOPE THAT IT WILL PROVIDE CUSTOMERS 
* WITH CODING INFORMATION TO SPEED UP THEIR PRODUCT DEVELOPMENT, BUT WITHOUT
* ANY WARRANTY; WITHOUT EVEN THE IMPLIED WARRANTY OF MERCHANTABILITY OR FITNESS
* FOR A PARTICULAR PURPOSE. AS A RESULT, TECHWELL SHALL NOT BE HELD LIABLE FOR ANY
* DIRECT, INDIRECT OR CONSEQUENTIAL DAMAGES WITH RESPECT TO ANY CLAIMS ARISING
* FROM THE CONTENT OF SUCH SOFTWARE AND/OR THE USE MADE BY CUSTOMERS OF THE
* CODING INFORMATION CONTAINED HEREIN IN CONNECTION WITH THEIR PRODUCTS.
*******************************************************************************/



#include <isil5864/tc_common.h>

//means nothing
int  ai_request_memory(struct file *f,unsigned long arg)
{
    int ret = 0;

    return ret;
}

//means nothing
int  ai_release_memory(struct file *f,unsigned long arg)
{
    return 0;
}



/*******************************************************************************
* Function Name  : tai_cleanup
* Description    : release function to release any resource it has.
* Input          : void *arg
* Output         : NULL
* Return         : NULL
*******************************************************************************/
void tai_cleanup(void * arg)
{
    struct isil_chip_ai *tai = (struct isil_chip_ai *)arg;

    kfree(tai);	
}




/*******************************************************************************
* Function Name  : chip_ai_open
* Description    : the open support function
* Input          : struct isil_service_device *tsd,struct file *f
* Output         : NULL
* Return         : 0:success; !0:failed
*******************************************************************************/
int chip_ai_open(struct isil_service_device *tsd,struct file *f)
{
    int ret = 0;
    ed_tcb_t   *aed; 
    struct isil_chip_ai *tai = container_of(tsd,struct isil_chip_ai,tsd);

    if(!tai->tsd.ped){
        dbg("ai dev has no driver attached!\n");
        return -1;
    }
    aed = tai->tsd.ped;
    ret = aed->op->open(aed);
    if(ret != ISIL_OK){
        ISIL_DBG(ISIL_DBG_INFO, ": audio_driver open failed!\n");
        ret = -1;
        return ret;
    }
    f->private_data = tai;
    return ret;
}


//implemented nothing right now
ssize_t chip_ai_read(struct file *f,char __user *buff,size_t count, loff_t * lof)
{
    int offset = 0;
    return (ssize_t)offset;
}


//implemented nothing right now
ssize_t chip_ai_write (struct file *f, const char __user *buff, size_t count, loff_t *lof)
{

    int ret = 0;

    return ret;

}

/*******************************************************************************
* Function Name  : chip_ai_ioctl
* Description    : ioctl support function
* Input          : struct file *f, unsigned int cmd, unsigned long arg
* Output         : NULL
* Return         : 0:success; !0:failed
*******************************************************************************/
int chip_ai_ioctl (struct file *f, unsigned int cmd, unsigned long arg)
{	
    int ret = 0;
    struct isil_chip_ai *tai = (struct isil_chip_ai *)f->private_data;
    struct endpoint_tcb *ed;
    //tc_trace;
    down(&tai->sem);
    switch(cmd)
    {
        case ISIL_CODEC_CHAN_REQUEST_MEM:
#ifdef METHOD_MAP_OLD
            ret = ai_request_memory(f,arg);
#endif
            break;
        case ISIL_CODEC_CHAN_RELEASE_MEM:
#ifdef METHOD_MAP_OLD
            ret = ai_release_memory(f,arg);
#endif
            break;
        default:
            //dbg("other cmd in %s\n",__FUNCTION__);
            if(!tai->tsd.ped){
                ISIL_DBG(ISIL_DBG_INFO, ": the device has no driver ed!\n");
                ret = EFAULT;
                break;
            }
            ed = tai->tsd.ped;
            ret = ed->op->ioctl(ed,cmd,arg);	
            break;
    }
    up(&tai->sem);
    return ret;
}


/*user close(fd) will call this function, it just perform close*/
/*******************************************************************************
* Function Name  : chip_ai_release
* Description    : release support function
* Input          : struct file *f
* Output         : NULL
* Return         : 0:success; !0:failed
*******************************************************************************/
int  chip_ai_release(struct file *f)
{
    int ret;
    ed_tcb_t   *aed;
    struct isil_chip_ai *tai = (struct isil_chip_ai *)f->private_data;

    tc_trace;
    down(&tai->sem);
    aed = tai->tsd.ped;
    ret = aed->op->close(aed);
    //tc_queue_reset(tai->tsd.queue);
    up(&tai->sem);

    f->private_data = NULL;
    return ret;
}




/*******************************************************************************
* Function Name  : chip_ai_init
* Description    : setup functions for ai
* Input          : struct isil_chip_device *tcd,struct cmd_arg *arg,struct isil_service_device **tsd
* Output         : NULL
* Return         : 0:success; !0:failed
*******************************************************************************/
int chip_ai_init(struct isil_chip_device *tcd,struct cmd_arg *arg,struct isil_service_device **tsd)
{
    struct isil_chip_ai *tai;

    tc_trace;
    tai = kzalloc(sizeof(*tai),GFP_KERNEL);
    if(NULL == tai)
        return -ENOMEM;
    *tsd = &tai->tsd;

    tai->tsd.clean_up = tai_cleanup;

    tai->name = "tw5864-audio_input";
    tai->users = 1;
    sema_init(&tai->sem, 1);
    //init_timer(&tai->polling_timer);
    init_waitqueue_head(&tai->wait_poll);
    return 0;
}

/*******************************************************************************
* Function Name  : chip_ai_destory
* Description    : destory the ai related tc_device
* Input          : struct isil_service_device *tsd
* Output         : NULL
* Return         : NULL
*******************************************************************************/
void chip_ai_destory(struct isil_service_device *tsd)
{
    struct isil_chip_ai *tai;
    tc_trace;

    if(!tsd)
        return;
    tai = container_of(tsd,struct isil_chip_ai,tsd);
    kfree(tai);
}


struct tsd_file_operations    chip_ai_fops = {
    .owner  = THIS_MODULE,
    .init   = chip_ai_init,
    .destory= chip_ai_destory,
    .open	= chip_ai_open,
    .read	= chip_ai_read,
    .write	= chip_ai_write,
    .ioctl  = chip_ai_ioctl,
    .release= chip_ai_release,
    .poll = NULL,
};

struct isil_dev_id  isil_chip_ai_id = {
    .epobj = {
		.vendor_id = ISIL,
#if	defined(PCI_PLATFORM)
        .bus_id =  ISIL_ED_PCI_BUS,
#else
        .bus_id =  ISIL_ED_HOST_BUS,
#endif
        .chip_id = TW5864,
        .func_id = 0,
        .type = ISIL_AUDIO_IN,
        .key = {0},
    },
    .version = 0,
};

static struct isil_dev_table_id *tai_table = NULL;

/*******************************************************************************
* Function Name  : isil_chip_ai_init
* Description    : register the ai dev_table_id to tc_core.
* Input          : NULL
* Output         : NULL
* Return         : 0:success; !0:failed
*******************************************************************************/
int isil_chip_ai_init(void)
{
    tai_table = kzalloc(sizeof(struct isil_dev_table_id),GFP_KERNEL);
    if(!tai_table){
        dbg("no mem in %s\n",__FUNCTION__);
        return -ENOMEM;	
    }
    INIT_LIST_HEAD(&tai_table->list);
    tai_table->tid = &isil_chip_ai_id;
    tai_table->ops = &chip_ai_fops;
    tai_table->para = NULL;//data;
    dbg_epobj_print(&tai_table->tid->epobj);
    register_tc_device(tai_table);
    return 0;	
}

/*******************************************************************************
* Function Name  : isil_chip_ai_remove
* Description    : unregister the ai dev_table_id from tc_core
* Input          : NULL
* Output         : NULL
* Return         : NULL
*******************************************************************************/
void isil_chip_ai_remove(void)
{
    if(tai_table)
        unregister_tc_device(tai_table);
    tai_table = NULL;
}

