/******************** (C) COPYRIGHT 2010 Intersil ********************
* File Name         : isil_push_buf_mgt.c
* Author            : rxchen@intersil.com
* Revision          : 
* Date              : 
* Description       : push buffer manager
*                   
* Change log        : 
*                   add file
********************************************************************************
* THE PRESENT SOFTWARE IS DISTRIBUTED IN THE HOPE THAT IT WILL PROVIDE CUSTOMERS 
* WITH CODING INFORMATION TO SPEED UP THEIR PRODUCT DEVELOPMENT, BUT WITHOUT
* ANY WARRANTY; WITHOUT EVEN THE IMPLIED WARRANTY OF MERCHANTABILITY OR FITNESS
* FOR A PARTICULAR PURPOSE. AS A RESULT, TECHWELL SHALL NOT BE HELD LIABLE FOR ANY
* DIRECT, INDIRECT OR CONSEQUENTIAL DAMAGES WITH RESPECT TO ANY CLAIMS ARISING
* FROM THE CONTENT OF SUCH SOFTWARE AND/OR THE USE MADE BY CUSTOMERS OF THE
* CODING INFORMATION CONTAINED HEREIN IN CONNECTION WITH THEIR PRODUCTS.
*******************************************************************************/

#include <isil5864/isil_common.h>



#if defined(PCI_PLATFORM)
#define ALIGN_ADDR  (PAGE_SIZE)
static __used int print_mem_info(isil_chip_t *chip)
{
    int i, j;
    push_buf_pool_t *pool;
    push_prev_buffer_t *buf;
    char arr[10];

    pool = &chip->push_buf_pool;

    ISIL_DBG(ISIL_DBG_INFO, "%8s %8s %8s %8s\n", "type", "index", "addr", "occupied");
    buf = &pool->preview_bufs[0];
    for (i = 0; i < PREVIEW_BUF_CNT; i++, buf++) {
        switch (buf->occupied) {
            case BUF_NO_OCCUPIED:
                strcpy(arr, "NO");
                break;
            case BUF_SD_OCCUPIED :
                strcpy(arr, "SD");
                break;
            case BUF_HD_OCCUPIED:
                strcpy(arr, "HD");
                break;
            default:
                strcpy(arr, "UNKNOWN");
        }
        ISIL_DBG(ISIL_DBG_INFO, "%8s %8d %8lx %8s\n", "PREVIEW", i, (unsigned long) buf->buf_addr, arr);

        if (buf->occupied == BUF_SD_OCCUPIED) {
            ISIL_DBG(ISIL_DBG_INFO, "%8s %8s %8s\n", "piece", "bitmap", "format");
            for (j = 0; j < PIECE_CNT_PER_BUF; j++) {
                switch (buf->fmt[j]) {
                    case ISIL_VIDEO_SIZE_D1:
                        strcpy(arr, "D1");
                        break;
                    case ISIL_VIDEO_SIZE_CIF:
                        strcpy(arr, "CIF");
                        break;
                    case ISIL_VIDEO_SIZE_QCIF:
                        strcpy(arr, "QCIF");
                        break;
                    case ISIL_VIDEO_SIZE_USER:
                        strcpy(arr, "USER");
                        break;	
                    default:
                        strcpy(arr, "UNKNOWN");
                }
                ISIL_DBG(ISIL_DBG_INFO, "%8d %8x %8s\n", j, buf->bit_map[j], arr);
            }
        }
    }

    return 0;
}


static ssize_t push_proc_read(struct seq_file *seq, void *data)
{
    isil_chip_t *chip;
    int i, j;
    push_buf_pool_t *pool;
    push_prev_buffer_t *buf;

    chip = (isil_chip_t *)seq->private;
    pool = &chip->push_buf_pool;

    seq_printf(seq, "%8s %8s %8s %8s %8s %8s\n", "type", "index", "vaddr", "paddr", "size", "occupied");
    buf = &pool->preview_bufs[0];
    for (i = 0; i < PREVIEW_BUF_CNT; i++, buf++) {
	    seq_printf(seq, "%8s %8d %lu %lu %8x", "PREVIEW", i, (unsigned long) buf->buf_addr, (unsigned long)buf->dma_addr, PREVIEW_BUF_SIZE);
        switch (buf->occupied) {
            case BUF_NO_OCCUPIED:
                seq_printf(seq, "%8s\n", "NO");
                break;
            case BUF_SD_OCCUPIED :
                seq_printf(seq, "%8s\n", "SD");
                break;
            case BUF_HD_OCCUPIED:
                seq_printf(seq, "%8s\n", "HD");
                break;
            default:
                seq_printf(seq, "%8s\n", "UNKNOWN");
        }

        if (buf->occupied == BUF_SD_OCCUPIED) {
            seq_printf(seq, "%8s %8s %8s\n", "piece", "bitmap", "format");
            for (j = 0; j < PIECE_CNT_PER_BUF; j++) {
                seq_printf(seq, "%8d %8x", j, buf->bit_map[j]);
                switch (buf->fmt[j]) {
                    case ISIL_VIDEO_SIZE_D1:
                        seq_printf(seq, "%8s\n", "D1");
                        break;
                    case ISIL_VIDEO_SIZE_CIF:
                        seq_printf(seq, "%8s\n", "CIF");
                        break;
                    case ISIL_VIDEO_SIZE_QCIF:
                        seq_printf(seq, "%8s\n", "QCIF");
                        break;
                    case ISIL_VIDEO_SIZE_USER:
                        seq_printf(seq, "%8s\n", "USER");
                        break;	
                    default:
                        seq_printf(seq, "%8s\n", "UNKNOWN");
                }
            }
        }
    }

    seq_printf(seq, "%8s %8s %8s %8s %8s %8s\n", "type", "index", "vaddr", "paddr", "size", "state");
    for (i = 0; i < H264_BUF_CNT; i++) {
        push_buffer_t *buf = &pool->h264_bufs[i];
        seq_printf(seq, "%8s %8d %lu %lu %8x %8x\n", "H264", i, (unsigned long) buf->buf_addr, (unsigned long)buf->dma_addr, H264_BUF_SIZE, buf->state);
    }

    for (i = 0; i < AUDIO_BUF_CNT; i++) {
        push_buffer_t *buf = &pool->audio_bufs[i];
        seq_printf(seq, "%8s %8d %lu %lu %8x %8x\n", "AUDIO", i, (unsigned long) buf->buf_addr, (unsigned long)buf->dma_addr, AUDIO_BUF_SIZE, buf->state);
    }

    for (i = 0; i < MV_BUF_CNT; i++) {
        push_buffer_t *buf = &pool->mv_bufs[i];
        seq_printf(seq, "%8s %8d %lu  %lu %8x %8x\n", "MV", i, (unsigned long) buf->buf_addr, (unsigned long)buf->dma_addr, MV_BUF_SIZE, buf->state);
    }

    for (i = 0; i < JPEG_BUF_CNT; i++) {
        push_buffer_t *buf = &pool->jpeg_bufs[i];
        seq_printf(seq, "%8s %8d %lu %lu %8x %8x\n", "MJPEG", i, (unsigned long) buf->buf_addr, (unsigned long)buf->dma_addr, JPEG_BUF_SIZE, buf->state);
    }

    return 0;
}

static ssize_t push_proc_write(struct file *file, const char __user *buffer,
        size_t count, loff_t *offset)
{
    return count;
}

typedef void (*push_buf_release_func)(push_buf_pool_t *, push_buffer_t *, int, int);



int find_order(int v, int max)
{
	int j;
	for (j = 0; j < max; j++) {
		v >>= 1;
		if (v == 0)
			break;
	}
	return j < max ? j : -1;
}

static void push_buf_release(push_buf_pool_t *pool, push_buffer_t *bufs, int buf_cnt, int buf_size)
{
	push_buffer_t *start_bufs;
	int i;
	start_bufs = bufs;
	for (i = 0; i < buf_cnt; i++, bufs++) {
#ifndef USE_DMA_ALLOC_COHERENT
		if(bufs->dma_addr){
			dma_unmap_single(pool->chip->dev, bufs->dma_addr, buf_size, DMA_FROM_DEVICE);
		}
		if (bufs->buf_addr){
			free_pages(bufs->buf_addr, buf_size);
		}
#else
		if (bufs->dma_addr){
			dma_free_coherent(pool->chip->dev, buf_size, bufs->buf_addr, bufs->dma_addr);
		}
#endif
		bufs->dma_addr = 0;
		bufs->buf_addr = 0;
		bufs->state = BUF_STATE_BUSY;
		bufs->cache_sync_len = ALIGN_ADDR;
	}
	return;
}

static int push_buf_init(push_buf_pool_t *pool, push_buffer_t *bufs,
			 struct spinlock *slock,
			 int buf_cnt, int buf_size,
			 int *free_cnt,
			 push_buf_release_func release)
{

	push_buffer_t *start_buf;
	int i;
	spin_lock_init(slock);
	start_buf = bufs;
	for (i = 0; i < buf_cnt; i++, bufs++) {
#ifndef USE_DMA_ALLOC_COHERENT
		int order;
		order = find_order(buf_size, 20);
		bufs->buf_addr = __get_free_pages(GFP_ATOMIC, order);
		if (bufs->buf_addr == NULL) {
			ISIL_DBG(ISIL_DBG_INFO, ": h264 can't get push buffer\n");
			goto out;
		}
		bufs->dma_addr = dma_map_single(pool->chip->dev, (void *)bufs->buf_addr, buf_size, DMA_BIDIRECTIONAL);
#else
		bufs->buf_addr = dma_alloc_coherent(pool->chip->dev, buf_size, &bufs->dma_addr, GFP_KERNEL);
#endif
		if(bufs->dma_addr == 0){
			ISIL_DBG(ISIL_DBG_INFO, ": h264 buf dma map err\n");
			goto out;
		}
		bufs->state = BUF_STATE_FREE;
		bufs->cache_sync_len = ALIGN_ADDR;
		*free_cnt = *free_cnt + 1;
	}
	return 0;
out:
	release(pool, start_buf, buf_cnt, buf_size);
	*free_cnt = 0;
	return -1;
}


static void push_preview_buf_release(push_buf_pool_t *pool,
				     push_prev_buffer_t *bufs,
				     int buf_cnt, int buf_size)
{
	push_prev_buffer_t *start_bufs;
	int i;
	start_bufs = bufs;

	for (i = 0; i < buf_cnt; i++, bufs++) {
#ifndef USE_DMA_ALLOC_COHERENT
		if(bufs->dma_addr){
			dma_unmap_single(pool->chip->dev, bufs->dma_addr, buf_size, DMA_FROM_DEVICE);
		}
		if (bufs->buf_addr){
			free_pages(bufs->buf_addr, buf_size);
		}
#else
		if (bufs->dma_addr){
			dma_free_coherent(pool->chip->dev, buf_size, bufs->buf_addr, bufs->dma_addr);
		}
#endif
		bufs->dma_addr = 0;
		bufs->buf_addr = 0;
		bufs->cache_sync_len = ALIGN_ADDR;
	}
	

	return;
}

static int push_preview_buf_init(push_buf_pool_t *pool,
				 push_prev_buffer_t *bufs,
				 struct spinlock *slock,
				 int buf_cnt, int buf_size,
				 int *free_cnt,
				 push_buf_release_func release)
{

	push_prev_buffer_t *start_buf;
	int i;
	int j;
	spin_lock_init(slock);
	start_buf = bufs;
	for (i = 0; i < buf_cnt; i++, bufs++) {
#ifndef USE_DMA_ALLOC_COHERENT
		int order;
		order = find_order(buf_size, 20);
		bufs->buf_addr = __get_free_pages(GFP_ATOMIC, order);
		if (bufs->buf_addr == NULL) {
			ISIL_DBG(ISIL_DBG_INFO, ": h264 can't get push buffer\n");
			goto out;
		}
		bufs->dma_addr = dma_map_single(pool->chip->dev, (void *)bufs->buf_addr, buf_size, DMA_BIDIRECTIONAL);
#else
		bufs->buf_addr = dma_alloc_coherent(pool->chip->dev, buf_size, &bufs->dma_addr, GFP_KERNEL);
#endif
		if(bufs->dma_addr == 0){
			ISIL_DBG(ISIL_DBG_INFO, ": h264 buf dma map err\n");
			goto out;
		}
		for (j = 0; j < PIECE_CNT_PER_BUF; j++) {
			bufs->bit_map[j] = 0;
			bufs->fmt[j] = ISIL_VIDEO_SIZE_USER;
		}
		bufs->occupied = BUF_NO_OCCUPIED;
		bufs->cache_sync_len = ALIGN_ADDR;
		*free_cnt = *free_cnt + 1;
	}
	*free_cnt = PIECE_CNT_PER_BUF * PREVIEW_BUF_CNT;
	return 0;
out:
	push_preview_buf_release(pool, start_buf, buf_cnt, buf_size);
	*free_cnt = 0;
	return -1;
}


int push_buf_pool_init(isil_chip_t *chip)
{
    push_buf_pool_t *pool = &chip->push_buf_pool;
    int ret = 0;
    isil_proc_register_s *push_buf_mgt_proc;

    memset(pool, 0, sizeof(push_buf_pool_t));
    pool->chip = chip;
    ret |= push_buf_init(pool, pool->h264_bufs, &pool->h264_queue_lock,
			 H264_BUF_CNT, H264_BUF_SIZE, &pool->h264_free_cnt,
			 push_buf_release);

    ret |= push_buf_init(pool, pool->audio_bufs, &pool->audio_queue_lock,
			 AUDIO_BUF_CNT, AUDIO_BUF_SIZE, &pool->audio_free_cnt,
			 push_buf_release);
    ret |= push_preview_buf_init(pool,
				 pool->preview_bufs, &pool->preview_queue_lock,
				 PREVIEW_BUF_CNT, PREVIEW_BUF_SIZE,
				 &pool->preview_free_cnt,
				 push_buf_release);
    ret |= push_buf_init(pool, pool->mv_bufs, &pool->mv_queue_lock,
			 MV_BUF_CNT, MV_BUF_SIZE, &pool->mv_free_cnt,
			 push_buf_release);
    ret |= push_buf_init(pool, pool->jpeg_bufs, &pool->jpeg_queue_lock,
			 JPEG_BUF_CNT, JPEG_BUF_SIZE, &pool->jpeg_free_cnt,
			 push_buf_release);

    push_buf_mgt_proc = &chip->push_buf_mgt_proc;
    if(!push_buf_mgt_proc->entry) {
        strcpy(push_buf_mgt_proc->name, "push_buf_mgt");
        push_buf_mgt_proc->read  = push_proc_read;
        push_buf_mgt_proc->write = push_proc_write;
        push_buf_mgt_proc->private = chip;
        isil_module_register(chip, push_buf_mgt_proc);
    }
    return ret;
}

void push_buf_pool_release(isil_chip_t *chip)
{
    push_buf_pool_t *pool = &chip->push_buf_pool;

    push_buf_release(pool, pool->h264_bufs, H264_BUF_CNT, H264_BUF_SIZE);
    push_buf_release(pool, pool->audio_bufs, AUDIO_BUF_CNT, AUDIO_BUF_SIZE);
    push_preview_buf_release(pool, pool->preview_bufs, PREVIEW_BUF_CNT,
			     PREVIEW_BUF_SIZE);
    push_buf_release(pool, pool->mv_bufs, MV_BUF_CNT, MV_BUF_SIZE);
    push_buf_release(pool, pool->jpeg_bufs, JPEG_BUF_CNT, JPEG_BUF_SIZE);

    pool->h264_free_cnt = 0;
    pool->audio_free_cnt = 0;
    pool->preview_free_cnt = 0;
    pool->mv_free_cnt = 0;
    pool->jpeg_free_cnt = 0;
    
    isil_module_unregister(chip, &chip->push_buf_mgt_proc);
}

int get_push_buf_cache_sync_len(int type, unsigned long addr, isil_chip_t *chip)
{
    push_buf_pool_t *pool = &chip->push_buf_pool;
    unsigned long lock_flags;
    int ret = ALIGN_ADDR, i;

    switch (type) {
        case BUF_TYPE_H264:
            spin_lock_irqsave(&pool->h264_queue_lock, lock_flags);
            for (i = 0; i < H264_BUF_CNT; i++) {
		    if ((unsigned long) pool->h264_bufs[i].buf_addr == addr){
                    ret = pool->h264_bufs[i].cache_sync_len;
                }
            }
            spin_unlock_irqrestore(&pool->h264_queue_lock, lock_flags);
            break;
       case BUF_TYPE_PREVIEW:
            ret = 0x20000;
            break;
        case BUF_TYPE_AUDIO:
            ret = 0x800;
            break;
        case BUF_TYPE_MV:
            spin_lock_irqsave(&pool->mv_queue_lock, lock_flags);
            for (i = 0; i < MV_BUF_CNT; i++) {
		    if ((unsigned long) pool->mv_bufs[i].buf_addr == addr){
                    ret = pool->mv_bufs[i].cache_sync_len;
                }
            }
            spin_unlock_irqrestore(&pool->mv_queue_lock, lock_flags);
            break;
        case BUF_TYPE_JPEG:
            spin_lock_irqsave(&pool->jpeg_queue_lock, lock_flags);
            for (i = 0; i < JPEG_BUF_CNT; i++) {
		    if ((unsigned long) pool->jpeg_bufs[i].buf_addr == addr){
			    ret = pool->jpeg_bufs[i].cache_sync_len;
                }
            }
            spin_unlock_irqrestore(&pool->jpeg_queue_lock, lock_flags);
            break;
        default:
            break;
    }
    return ret;
}

void    update_push_buf_cache_sync_len(int type, int sync_len, unsigned long addr, isil_chip_t *chip)
{
    push_buf_pool_t *pool = &chip->push_buf_pool;
    unsigned long lock_flags;
    int i;

	sync_len = PAGE_ALIGN(sync_len + ALIGN_ADDR);
    switch (type) {
        case BUF_TYPE_H264:
            spin_lock_irqsave(&pool->h264_queue_lock, lock_flags);
            for (i = 0; i < H264_BUF_CNT; i++) {
		    if ((unsigned long) pool->h264_bufs[i].buf_addr == addr){
                    pool->h264_bufs[i].cache_sync_len = sync_len;
                }
            }
            spin_unlock_irqrestore(&pool->h264_queue_lock, lock_flags);
            break;
        case BUF_TYPE_PREVIEW:
            break;
        case BUF_TYPE_AUDIO:
            break;
        case BUF_TYPE_MV:
            spin_lock_irqsave(&pool->mv_queue_lock, lock_flags);
            for (i = 0; i < MV_BUF_CNT; i++) {
		    if ((unsigned long) pool->mv_bufs[i].buf_addr == addr){
			    pool->mv_bufs[i].cache_sync_len = sync_len;
                }
            }
            spin_unlock_irqrestore(&pool->mv_queue_lock, lock_flags);
            break;
        case BUF_TYPE_JPEG:
            spin_lock_irqsave(&pool->jpeg_queue_lock, lock_flags);
            for (i = 0; i < JPEG_BUF_CNT; i++) {
		    if ((unsigned long) pool->jpeg_bufs[i].buf_addr == addr){
			    pool->jpeg_bufs[i].cache_sync_len = sync_len;
                }
            }
            spin_unlock_irqrestore(&pool->jpeg_queue_lock, lock_flags);
            break;
        default:
            break;
    }
}

unsigned long	get_push_buf(int type, int fmt, dma_addr_t *dma_addr, isil_chip_t *chip)
{
    push_buf_pool_t *pool = &chip->push_buf_pool;
    unsigned long lock_flags;
    unsigned long addr, offset;
    push_prev_buffer_t *buf;
    int i, j, k;

	addr = 0;
	*dma_addr = 0;
    switch (type) {
        case BUF_TYPE_H264:
            spin_lock_irqsave(&pool->h264_queue_lock, lock_flags);
            if (pool->h264_free_cnt == 0) {
                ISIL_DBG(ISIL_DBG_INFO, ": no free h264 buf.\n");
                spin_unlock_irqrestore(&pool->h264_queue_lock, lock_flags);
                goto err;
            }

            for (i = 0; i < H264_BUF_CNT; i++) {
                if (pool->h264_bufs[i].state == BUF_STATE_FREE)
                    break;
            }
            if (i == H264_BUF_CNT) {
                ISIL_DBG(ISIL_DBG_INFO, ": no free h264 buf, free_cnt is wrong.\n");
                spin_unlock_irqrestore(&pool->h264_queue_lock, lock_flags);
                goto err;
            }

            pool->h264_bufs[i].state = BUF_STATE_BUSY;
            pool->h264_free_cnt--;
            addr = (unsigned long) pool->h264_bufs[i].buf_addr;
            *dma_addr = pool->h264_bufs[i].dma_addr;
            spin_unlock_irqrestore(&pool->h264_queue_lock, lock_flags);
            return addr;

        case BUF_TYPE_PREVIEW:
            spin_lock_irqsave(&pool->preview_queue_lock, lock_flags);

            buf = pool->preview_bufs;
            for (i = 0; i < PREVIEW_BUF_CNT; i++, buf++) {
                if (buf->occupied == BUF_SD_OCCUPIED) {
                    for (j = 0; j < PIECE_CNT_PER_BUF; j++) {
                        if (buf->fmt[j] == fmt) {
                            switch (fmt) {
                                case ISIL_VIDEO_SIZE_D1:
                                    break;

                                case ISIL_VIDEO_SIZE_CIF:
                                    for (k = 0; k < CIF_CNT_PER_PIECE; k++) {
                                        if (!(buf->bit_map[j] & (0xf << 4*k))) {
                                            buf->bit_map[j] |= 0xf << 4*k;
					    offset = j*PIECE_PUSH_SIZE + k*CIF_PUSH_SIZE;
					    addr = (unsigned long) buf->buf_addr + offset;
                                            *dma_addr = buf->dma_addr + offset;
                                            spin_unlock_irqrestore(&pool->preview_queue_lock, lock_flags);
                                            return addr;
                                        }
                                    }
                                    break;

                                case ISIL_VIDEO_SIZE_QCIF:
                                    for (k = 0; k < QCIF_CNT_PER_PIECE; k++) {
                                        if (!(buf->bit_map[j] & (0x1 << k))) {
                                            buf->bit_map[j] |= 0x1 << k;
					    offset = j*PIECE_PUSH_SIZE + k*QCIF_PUSH_SIZE;
					    addr = (unsigned long)  buf->buf_addr + offset;
                                            *dma_addr = buf->dma_addr + offset;
                                            spin_unlock_irqrestore(&pool->preview_queue_lock, lock_flags);
                                            return addr;
                                        }
                                    }
                                    break;

                                default:
                                    ISIL_DBG(ISIL_DBG_INFO, ": invalid format type %d\n", fmt);
                                    spin_unlock_irqrestore(&pool->preview_queue_lock, lock_flags);
                                    goto err;
                            }				
                        }
                    }
                }
            }

            buf = pool->preview_bufs;
            for (i = 0; i < PREVIEW_BUF_CNT; i++, buf++) {
                if (buf->occupied == BUF_SD_OCCUPIED) {
                    for (j = 0; j < PIECE_CNT_PER_BUF; j++) {
						offset = j*PIECE_PUSH_SIZE;
                        if (buf->fmt[j] == ISIL_VIDEO_SIZE_USER) {
                            switch (fmt) {
                                case ISIL_VIDEO_SIZE_D1:
                                    buf->fmt[j] = fmt;
                                    buf->bit_map[j] = 0xffff;
                                    pool->preview_free_cnt--;
                                    addr = (unsigned long) buf->buf_addr + offset;
                                    *dma_addr = buf->dma_addr + offset;
                                    spin_unlock_irqrestore(&pool->preview_queue_lock, lock_flags);
                                    return addr;

                                case ISIL_VIDEO_SIZE_CIF:
                                    buf->fmt[j] = fmt;
                                    buf->bit_map[j] = 0xf;
                                    pool->preview_free_cnt--;
                                    addr = (unsigned long) buf->buf_addr + offset;
                                    *dma_addr = buf->dma_addr + offset;
                                    spin_unlock_irqrestore(&pool->preview_queue_lock, lock_flags);
                                    return addr;

                                case ISIL_VIDEO_SIZE_QCIF:
                                    buf->fmt[j] = fmt;
                                    buf->bit_map[j] = 0x1;
                                    pool->preview_free_cnt--;
                                    addr = ((unsigned long) buf->buf_addr + offset);
                                    *dma_addr = buf->dma_addr + offset;
                                    spin_unlock_irqrestore(&pool->preview_queue_lock, lock_flags);
                                    return addr;

                                default:
                                    ISIL_DBG(ISIL_DBG_INFO, ": invalid format type %d\n", fmt);
                                    spin_unlock_irqrestore(&pool->preview_queue_lock, lock_flags);
                                    goto err;
                            }				
                        }
                    }
                }
            }

            buf = pool->preview_bufs;
            for (i = 0; i < PREVIEW_BUF_CNT; i++, buf++) {
                if (buf->occupied == BUF_NO_OCCUPIED) {
                    buf->occupied = BUF_SD_OCCUPIED;
                    switch (fmt) {
                        case ISIL_VIDEO_SIZE_D1:
                            buf->fmt[0] = fmt;
                            buf->bit_map[0] = 0xffff;
                            pool->preview_free_cnt--;
                            addr = (unsigned long) buf->buf_addr;
                            *dma_addr = buf->dma_addr;
                            spin_unlock_irqrestore(&pool->preview_queue_lock, lock_flags);
                            return addr;

                        case ISIL_VIDEO_SIZE_CIF:
                            buf->fmt[0] = fmt;
                            buf->bit_map[0] = 0xf;
                            pool->preview_free_cnt--;
                            addr = (unsigned long) buf->buf_addr;
                            *dma_addr = buf->dma_addr;
                            spin_unlock_irqrestore(&pool->preview_queue_lock, lock_flags);
                            return addr;

                        case ISIL_VIDEO_SIZE_QCIF:
                            buf->fmt[0] = fmt;
                            buf->bit_map[0] = 0x1;
                            pool->preview_free_cnt--;
                            addr = (unsigned long) buf->buf_addr;
                            *dma_addr = buf->dma_addr;
                            spin_unlock_irqrestore(&pool->preview_queue_lock, lock_flags);
                            return addr;

                        default:
                            ISIL_DBG(ISIL_DBG_INFO, ": invalid format type %d\n", fmt);
                            spin_unlock_irqrestore(&pool->preview_queue_lock, lock_flags);
                            goto err;
                    }				
                }
            }
            spin_unlock_irqrestore(&pool->preview_queue_lock, lock_flags);
            goto err;

        case BUF_TYPE_AUDIO:
            spin_lock_irqsave(&pool->audio_queue_lock, lock_flags);
            if (pool->audio_free_cnt == 0) {
                ISIL_DBG(ISIL_DBG_INFO, ": no free audio buf.\n");
                spin_unlock_irqrestore(&pool->audio_queue_lock, lock_flags);
                goto err;
            }

            for (i = 0; i < AUDIO_BUF_CNT; i++) {
                if (pool->audio_bufs[i].state == BUF_STATE_FREE)
                    break;
            }
            if (i == AUDIO_BUF_CNT) {
                ISIL_DBG(ISIL_DBG_INFO, ": no free h264 buf, free_cnt is wrong.\n");
                spin_unlock_irqrestore(&pool->audio_queue_lock, lock_flags);
                goto err;
            }

            pool->audio_bufs[i].state = BUF_STATE_BUSY;
            pool->audio_free_cnt--;
            addr = (unsigned long) pool->audio_bufs[i].buf_addr;
            *dma_addr = pool->audio_bufs[i].dma_addr;
            spin_unlock_irqrestore(&pool->audio_queue_lock, lock_flags);
            return addr;

        case BUF_TYPE_MV:
            spin_lock_irqsave(&pool->mv_queue_lock, lock_flags);
            if (pool->mv_free_cnt == 0) {
                ISIL_DBG(ISIL_DBG_INFO, ": no free mv buf.\n");
                spin_unlock_irqrestore(&pool->mv_queue_lock, lock_flags);
                goto err;
            }

            for (i = 0; i < MV_BUF_CNT; i++) {
                if (pool->mv_bufs[i].state == BUF_STATE_FREE)
                    break;
            }
            if (i == MV_BUF_CNT) {
                ISIL_DBG(ISIL_DBG_INFO, ": no free mv buf, free_cnt is wrong.\n");
                spin_unlock_irqrestore(&pool->mv_queue_lock, lock_flags);
                goto err;
            }

            pool->mv_bufs[i].state = BUF_STATE_BUSY;
            pool->mv_free_cnt--;
            addr = (unsigned long) pool->mv_bufs[i].buf_addr;
            *dma_addr = pool->mv_bufs[i].dma_addr;
            spin_unlock_irqrestore(&pool->mv_queue_lock, lock_flags);
            return addr;

        case BUF_TYPE_JPEG:
            spin_lock_irqsave(&pool->jpeg_queue_lock, lock_flags);
            if (pool->jpeg_free_cnt == 0) {
                ISIL_DBG(ISIL_DBG_INFO, ": no free jpeg buf.\n");
                spin_unlock_irqrestore(&pool->jpeg_queue_lock, lock_flags);
                goto err;
            }

            for (i = 0; i < JPEG_BUF_CNT; i++) {
                if (pool->jpeg_bufs[i].state == BUF_STATE_FREE)
                    break;
            }
            if (i == JPEG_BUF_CNT) {
                ISIL_DBG(ISIL_DBG_INFO, ": no free jpeg buf, free_cnt is wrong.\n");
                spin_unlock_irqrestore(&pool->jpeg_queue_lock, lock_flags);
                goto err;
            }

            pool->jpeg_bufs[i].state = BUF_STATE_BUSY;
            pool->jpeg_free_cnt--;
            addr = (unsigned long) pool->jpeg_bufs[i].buf_addr;
            *dma_addr = pool->jpeg_bufs[i].dma_addr;
            spin_unlock_irqrestore(&pool->jpeg_queue_lock, lock_flags);
            return addr;

        default:
            ISIL_DBG(ISIL_DBG_INFO, ": get push buf, type err\n");
            goto err;
    }

err:
    return addr;
}

static int buf_occupy_empty(u32 *bmap, int cnt)
{
    int i;
    int ret = 0;

    for (i = 0; i < cnt; i++) {
        ret |= bmap[i];
    }
    return !ret;
}

void release_push_buf(int type, int fmt, unsigned long addr, isil_chip_t *chip)
{
    push_buf_pool_t *pool = &chip->push_buf_pool;
    unsigned long lock_flags;
    push_prev_buffer_t *buf;
    int i, j, k;

    switch (type) {
        case BUF_TYPE_H264:
            spin_lock_irqsave(&pool->h264_queue_lock, lock_flags);
            for (i = 0; i < H264_BUF_CNT; i++) {
		    if ((unsigned long) pool->h264_bufs[i].buf_addr == addr)
			    break;
            }
            if (i == H264_BUF_CNT) {
                ISIL_DBG(ISIL_DBG_INFO, ": can't find right h264 buf %lu, check it out.\n", addr);
                spin_unlock_irqrestore(&pool->h264_queue_lock, lock_flags);
                return;
            }			

            if (pool->h264_bufs[i].state != BUF_STATE_BUSY)
                ISIL_DBG(ISIL_DBG_INFO, ": buf state isn't right, check it out.\n");
            pool->h264_bufs[i].state = BUF_STATE_FREE;
            pool->h264_free_cnt++;
            spin_unlock_irqrestore(&pool->h264_queue_lock, lock_flags);
            break;

        case BUF_TYPE_PREVIEW:
            spin_lock_irqsave(&pool->preview_queue_lock, lock_flags);

            buf = pool->preview_bufs;
            for (i = 0; i < PREVIEW_BUF_CNT; i++, buf++) {
                if (buf->occupied == BUF_SD_OCCUPIED) {
                    for (j = 0; j < PIECE_CNT_PER_BUF; j++) {
                        if (buf->fmt[j] == fmt) {
                            switch (fmt) {
                                case ISIL_VIDEO_SIZE_D1:
					if (((unsigned long) buf->buf_addr + j*PIECE_PUSH_SIZE) == addr) {
                                        if (buf->bit_map[j] != 0xffff)
                                            ISIL_DBG(ISIL_DBG_INFO, ": buf bit_map %x isn't right, check it out.\n", buf->bit_map[j]);
                                        buf->fmt[j] = ISIL_VIDEO_SIZE_USER;
                                        buf->bit_map[j] = 0;
                                        pool->preview_free_cnt++;
                                        if (buf_occupy_empty(buf->bit_map, PIECE_CNT_PER_BUF))
                                            buf->occupied = BUF_NO_OCCUPIED;
                                        spin_unlock_irqrestore(&pool->preview_queue_lock, lock_flags);
                                        return;
                                    }				
                                    break;

                                case ISIL_VIDEO_SIZE_CIF:
                                    for (k = 0; k < CIF_CNT_PER_PIECE; k++) {
					    if ((unsigned long) ((unsigned long) buf->buf_addr + j*PIECE_PUSH_SIZE + k*CIF_PUSH_SIZE) == addr)
						    break;
                                    }
                                    if (k == CIF_CNT_PER_PIECE)
                                        break;

                                    if (!(buf->bit_map[j] & (0xf << 4*k)))
                                        ISIL_DBG(ISIL_DBG_INFO, ": buf bit_map %x isn't right, k=%d.\n", buf->bit_map[j], k);
                                    buf->bit_map[j] &= ~(0xf << 4*k);
                                    if (!(buf->bit_map[j] & MASK_LOW16)) {
                                        buf->fmt[j] = ISIL_VIDEO_SIZE_USER;
                                        buf->bit_map[j] = 0;
                                        pool->preview_free_cnt++;
                                        if (buf_occupy_empty(buf->bit_map, PIECE_CNT_PER_BUF))
                                            buf->occupied = BUF_NO_OCCUPIED;
                                    }
                                    spin_unlock_irqrestore(&pool->preview_queue_lock, lock_flags);
                                    return;

                                case ISIL_VIDEO_SIZE_QCIF:
                                    for (k = 0; k < QCIF_CNT_PER_PIECE; k++) {
					    if ((unsigned long) buf->buf_addr + j*PIECE_PUSH_SIZE + k*QCIF_PUSH_SIZE == addr)
                                            break;
                                    }
                                    if (k == QCIF_CNT_PER_PIECE)
                                        break;

                                    if (!(buf->bit_map[j] & (0x1 << k)))
                                        ISIL_DBG(ISIL_DBG_INFO, ": buf bit_map %x isn't right, k=%d.\n", buf->bit_map[j], k);
                                    buf->bit_map[j] &= ~(0x1 << k);
                                    if (!(buf->bit_map[j] & MASK_LOW16)) {
                                        buf->fmt[j] = ISIL_VIDEO_SIZE_USER;
                                        buf->bit_map[j] = 0;
                                        pool->preview_free_cnt++;
                                        if (buf_occupy_empty(buf->bit_map, PIECE_CNT_PER_BUF))
                                            buf->occupied = BUF_NO_OCCUPIED;
                                    }
                                    spin_unlock_irqrestore(&pool->preview_queue_lock, lock_flags);
                                    return;

                                default:
                                    ISIL_DBG(ISIL_DBG_INFO, ": invalid format type %d\n", fmt);
                                    spin_unlock_irqrestore(&pool->preview_queue_lock, lock_flags);
                                    return;
                            }				
                        }
                    }
                }
            }
            ISIL_DBG(ISIL_DBG_INFO, ": can't find right fmt %d buf %lu, check it out.\n", fmt, addr);
            spin_unlock_irqrestore(&pool->preview_queue_lock, lock_flags);
            break;

        case BUF_TYPE_AUDIO:
            spin_lock_irqsave(&pool->audio_queue_lock, lock_flags);
            for (i = 0; i < AUDIO_BUF_CNT; i++) {
		    if ((unsigned long) pool->audio_bufs[i].buf_addr == (unsigned long) addr)
			    break;
            }
            if (i == AUDIO_BUF_CNT) {
                ISIL_DBG(ISIL_DBG_INFO, ": can't find right audio buf %lu, check it out.\n", addr);
                spin_unlock_irqrestore(&pool->audio_queue_lock, lock_flags);
                return;
            }			

            if (pool->audio_bufs[i].state != BUF_STATE_BUSY)
                ISIL_DBG(ISIL_DBG_INFO, ": buf state isn't right, check it out.\n");
            pool->audio_bufs[i].state = BUF_STATE_FREE;
            pool->audio_free_cnt++;
            spin_unlock_irqrestore(&pool->audio_queue_lock, lock_flags);
            break;

        case BUF_TYPE_MV:
            spin_lock_irqsave(&pool->mv_queue_lock, lock_flags);
            for (i = 0; i < MV_BUF_CNT; i++) {
		    if ((unsigned long) pool->mv_bufs[i].buf_addr == (unsigned long) addr)
			    break;
            }
            if (i == MV_BUF_CNT) {
                ISIL_DBG(ISIL_DBG_INFO, ": can't find right mv buf %lu, check it out.\n", addr);
                spin_unlock_irqrestore(&pool->mv_queue_lock, lock_flags);
                return;
            }			

            if (pool->mv_bufs[i].state != BUF_STATE_BUSY)
                ISIL_DBG(ISIL_DBG_INFO, ": buf state isn't right, check it out.\n");
            pool->mv_bufs[i].state = BUF_STATE_FREE;
            pool->mv_free_cnt++;
            spin_unlock_irqrestore(&pool->mv_queue_lock, lock_flags);
            break;

        case BUF_TYPE_JPEG:
            spin_lock_irqsave(&pool->jpeg_queue_lock, lock_flags);
            for (i = 0; i < JPEG_BUF_CNT; i++) {
		    if ((unsigned long) pool->jpeg_bufs[i].buf_addr == (unsigned long) addr)
			    break;
            }
            if (i == JPEG_BUF_CNT) {
                ISIL_DBG(ISIL_DBG_INFO, ": can't find right jpeg buf %lu, check it out.\n", addr);
                spin_unlock_irqrestore(&pool->jpeg_queue_lock, lock_flags);
                return;
            }			

            if (pool->jpeg_bufs[i].state != BUF_STATE_BUSY)
                ISIL_DBG(ISIL_DBG_INFO, ": jpeg buf state isn't right, check it out.\n");
            pool->jpeg_bufs[i].state = BUF_STATE_FREE;
            pool->jpeg_free_cnt++;
            spin_unlock_irqrestore(&pool->jpeg_queue_lock, lock_flags);
            break;

        default:
            ISIL_DBG(ISIL_DBG_INFO, ": get push buf, type err\n");
    }
}
#endif

