/***************************************** (C) COPYRIGHT 2010 Intersil *****************************************
* File Name         : isil_tasklet.c
* Author            : tangt.sc@intersil.com
* Revision          : 1.2.1.1275
* Date              : 2012-03-29
* Description       : process audio&h264 encode bitstream copy
* Change log        : 
*                   
****************************************************************************************************************
* THE PRESENT SOFTWARE IS DISTRIBUTED IN THE HOPE THAT IT WILL PROVIDE CUSTOMERS 
* WITH CODING INFORMATION TO SPEED UP THEIR PRODUCT DEVELOPMENT, BUT WITHOUT
* ANY WARRANTY; WITHOUT EVEN THE IMPLIED WARRANTY OF MERCHANTABILITY OR FITNESS
* FOR A PARTICULAR PURPOSE. AS A RESULT, TECHWELL SHALL NOT BE HELD LIABLE FOR ANY
* DIRECT, INDIRECT OR CONSEQUENTIAL DAMAGES WITH RESPECT TO ANY CLAIMS ARISING
* FROM THE CONTENT OF SUCH SOFTWARE AND/OR THE USE MADE BY CUSTOMERS OF THE
* CODING INFORMATION CONTAINED HEREIN IN CONNECTION WITH THEIR PRODUCTS.
****************************************************************************************************************/
#include	<isil5864/isil_common.h>

#if defined(PCI_PLATFORM)
/****************************************************************************************************************
* Function Name :   isil_tasklet_init
* Description   :   init one tasklet
* Input         :   param1 is management tasklet's entry
*                   param2 is callback function ptr
*                   param3 is param
* Output        :   null
* Return        :   null
****************************************************************************************************************/
static void isil_tasklet_init(isil_tasklet_entry_t *isil_tasklet_entry, isil_tasklet_hook callfn, unsigned long data)
{
    tasklet_init(&isil_tasklet_entry->task, callfn, data);
}

/****************************************************************************************************************
* Function Name :   isil_tasklet_close
* Description   :   close one tasklet
* Input         :   param1 is management tasklet's entry
* Output        :   null
* Return        :   null
****************************************************************************************************************/
static void isil_tasklet_close(isil_tasklet_entry_t *isil_tasklet_entry)
{
    tasklet_kill(&isil_tasklet_entry->task);
}

/****************************************************************************************************************
* Function Name :   isil_tasklet_enable
* Description   :   enable one tasklet
* Input         :   param1 is management tasklet's entry
* Output        :   null
* Return        :   null
****************************************************************************************************************/
static void isil_tasklet_enable(isil_tasklet_entry_t *isil_tasklet_entry)
{
    tasklet_enable(&isil_tasklet_entry->task);
}

/****************************************************************************************************************
* Function Name :   isil_tasklet_disable
* Description   :   disable one tasklet
* Input         :   param1 is management tasklet's entry
* Output        :   null
* Return        :   null
****************************************************************************************************************/
static void isil_tasklet_disable(isil_tasklet_entry_t *isil_tasklet_entry)
{
    tasklet_disable(&isil_tasklet_entry->task);
}

/****************************************************************************************************************
* Function Name :   isil_tasklet_schedule
* Description   :   triggle tasklet schedule
* Input         :   param1 is management tasklet's entry
* Output        :   null
* Return        :   null
****************************************************************************************************************/
static void isil_tasklet_schedule(isil_tasklet_entry_t *isil_tasklet_entry)
{
    tasklet_schedule(&isil_tasklet_entry->task);
}

/****************************************************************************************************************
* Veriable Name :   isil_tasklet_op
* Description   :   for helping op tasklet, we define general interface
****************************************************************************************************************/
static struct isil_tasklet_operation    isil_tasklet_op = {
    .init = isil_tasklet_init,
    .close = isil_tasklet_close,
    .enable = isil_tasklet_enable,
    .disable = isil_tasklet_disable,
    .schedule = isil_tasklet_schedule,
};

/****************************************************************************************************************
* Function Name :   h264_encode_tasklet
* Description   :   h264 copy bitstream function
* Input         :   param1 is register param
* Output        :   null
* Return        :   null
****************************************************************************************************************/
static void h264_encode_tasklet(unsigned long data)
{
    isil_h264_logic_encode_chan_t   *h264_logic_encode_chan;
    isil_tasklet_entry_t            *encode_tasklet;
    ed_tcb_t                        *opened_logic_chan_ed;
    isil_h264_encode_control_t      *h264_encode_control;
    isil_video_frame_tcb_queue_t    *video_frame_queue;
    isil_video_frame_tcb_t          *frame;
    isil_video_packet_tcb_queue_t   *video_packet_queue;
    isil_video_chan_buf_pool_t      *video_chan_buf_pool;
    isil_video_packet_tcb_t         *packet;
    isil_chip_t *chip;
    h264_nal_t  *nal;
    u32 encode_len;
    u8  temp_encode_data;
	unsigned long	flags;

    h264_logic_encode_chan = (isil_h264_logic_encode_chan_t*)data;
    encode_tasklet = &h264_logic_encode_chan->encode_tasklet;
    opened_logic_chan_ed = &h264_logic_encode_chan->opened_logic_chan_ed;
    h264_encode_control = &h264_logic_encode_chan->encode_control;
    video_frame_queue = &h264_logic_encode_chan->encode_frame_queue;
    video_chan_buf_pool = &h264_logic_encode_chan->encode_chan_buf_pool;
    chip = h264_logic_encode_chan->chip;

    frame = video_frame_queue->curr_producer;
    if(frame){
        video_packet_queue = &frame->video_packet_queue;
        packet = video_packet_queue->curr_producer;
        nal = &frame->nal;
        encode_len = nal->encode_len;

        if(frame->frame_is_err==0 && packet){
            u8  *src_ptr, *dest_ptr;
            int transfer_state;

            pci_dma_sync_single_for_cpu(chip->pci_dev, h264_logic_encode_chan->h264_push_dma_addr, get_push_buf_cache_sync_len(BUF_TYPE_H264, h264_logic_encode_chan->push_buf, chip), DMA_FROM_DEVICE);
			update_push_buf_cache_sync_len(BUF_TYPE_H264, encode_len, h264_logic_encode_chan->push_buf, chip);
            src_ptr = (unsigned char *)h264_logic_encode_chan->push_buf;
            temp_encode_data = *src_ptr++;
            encode_len--;
            temp_encode_data <<= (nal->i_temp_bitalign_number);
            temp_encode_data >>= (nal->i_temp_bitalign_number);
            temp_encode_data |= (u8)(nal->i_temp_bitalign_content);
            packet->data[frame->frame_len] = temp_encode_data;
            frame->frame_len++;
            packet->payload_len = frame->frame_len;
            transfer_state = 0;
            while(encode_len){
                dest_ptr = &packet->data[packet->payload_len];
                while((packet->payload_len < packet->packet_size) && (encode_len > 0)){
                    if(transfer_state == 2 && (src_ptr[0]<=0x03)){
                        *dest_ptr++ = 0x03;
                        packet->payload_len++;
                        frame->frame_len++;
                        transfer_state = 0;
                        if(packet->payload_len == packet->packet_size){
                            break;
                        }
                    }
                    if(src_ptr[0] == 0) {
                        if(transfer_state < 2) {
                            transfer_state++;
                        }
                    } else {
                        transfer_state = 0;
                    }
                    *dest_ptr++ = *src_ptr++;
                    packet->payload_len++;
                    frame->frame_len++;
                    encode_len--;
                }

                video_packet_queue->op->put_curr_producer_into_queue(video_packet_queue);
                if(encode_len){
                    video_packet_queue->op->try_get_curr_producer_from_pool(video_packet_queue, &h264_logic_encode_chan->encode_chan_buf_pool);
                    packet = video_packet_queue->curr_producer;
                    if (packet == NULL) {
                        frame->frame_is_err = 1;
                    }
                }
				if(frame->frame_is_err){
                    h264_encode_control->op->update_slice_head_discard(h264_encode_control);
                    ISIL_DBG(ISIL_DBG_INFO, "can't get packet tcb, len=%d\n", encode_len);
                    break;
				}
            }
        } else {
			frame->frame_is_err = 1;
        }

        if(frame->frame_is_err){
            video_frame_queue->op->release_curr_producer(video_frame_queue, video_chan_buf_pool);
            h264_encode_control->op->force_I_frame(h264_encode_control);
            ISIL_DBG(ISIL_DBG_INFO, "discard, force I, %d.%d.%d\n", h264_logic_encode_chan->phy_slot_id, h264_logic_encode_chan->logic_chan_id, atomic_read(&h264_logic_encode_chan->opened_flag));
        } else {
            video_frame_queue->op->put_curr_producer_into_queue(video_frame_queue);
        }
    } else {
        h264_encode_control->op->force_I_frame(h264_encode_control);
        ISIL_DBG(ISIL_DBG_INFO, "discard, force I, %d.%d.%d\n", h264_logic_encode_chan->phy_slot_id, h264_logic_encode_chan->logic_chan_id, atomic_read(&h264_logic_encode_chan->opened_flag));
    }

    if (h264_logic_encode_chan->push_buf) {
        release_push_buf(BUF_TYPE_H264, 0, h264_logic_encode_chan->push_buf, chip);
        h264_logic_encode_chan->push_buf = 0;
    }
    if (h264_logic_encode_chan->mv_push_buf) {
        release_push_buf(BUF_TYPE_MV, 0, h264_logic_encode_chan->mv_push_buf, chip);
        h264_logic_encode_chan->mv_push_buf = 0;
    }

    spin_lock_irqsave(&h264_logic_encode_chan->fsm_matrix_call_lock, flags);
    driver_gen_deliver_event(opened_logic_chan_ed, 1);  //transfering to done.
    spin_unlock_irqrestore(&h264_logic_encode_chan->fsm_matrix_call_lock, flags);        

    isil_h264_logic_encode_chan_monitor(h264_logic_encode_chan, h264_logic_encode_chan->master_slot);
}

/****************************************************************************************************************
* Function Name :   init_isil_h264_encode_tasklet
* Description   :   init one tasklet for copying 264 bitstram
* Input         :   param1 is tasklet entry
*                   param2 is register param
* Output        :   null
* Return        :   null
****************************************************************************************************************/
void    init_isil_h264_encode_tasklet(isil_tasklet_entry_t *isil_tasklet_entry, unsigned long data)
{
    isil_tasklet_entry->op = &isil_tasklet_op;
    isil_tasklet_entry->op->init(isil_tasklet_entry, h264_encode_tasklet, data);
}

/****************************************************************************************************************
* Function Name :   close_isil_h264_encode_tasklet
* Description   :   close for h264's tasklet
* Input         :   param1 is tasklet entry
* Output        :   null
* Return        :   null
****************************************************************************************************************/
void    close_isil_h264_encode_tasklet(isil_tasklet_entry_t *isil_tasklet_entry)
{
    if(isil_tasklet_entry->op){
        isil_tasklet_entry->op->close(isil_tasklet_entry);
    }
}

/****************************************************************************************************************
* Function Name :   audio_encode_tasklet
* Description   :   audio tasklet's callback
* Input         :   param1 is tasklet entry
*                   param2 is register param
* Output        :   null
* Return        :   null
****************************************************************************************************************/
static void audio_encode_tasklet(unsigned long data)
{
    isil_chip_t		*chip;
    chip_audio_t	*chip_audio;
    chip_audio_push_isr_interface_t *audio_push_isr_interface;
    chip_audio_push_buf_state_t     *audio_push_buf;
    isil_audio_driver_t	*audio_chan_driver;
    ed_tcb_t            *ed_tcb;
    isil_audio_chan_push_interface_t  *audio_chan_push_interface;
    unsigned int    chan_id_flags, chan_id, rd_ptr;
    char        *curr_push_buf_addr, *src_buf;
    dma_addr_t  curr_dma_addr;

    chip = (isil_chip_t*)data;
    chip_audio = &chip->chip_audio;
    audio_push_isr_interface = &chip_audio->audio_push_interface;
    audio_push_buf = &audio_push_isr_interface->audio_push_buf;

    for(chan_id=0; chan_id<16; chan_id++){
        while(!audio_push_buf->op->push_buf_is_empty(audio_push_buf, chan_id)){
            rd_ptr = audio_push_buf->op->get_push_buf_rd_ptr(audio_push_buf, chan_id);
            chan_id_flags = audio_push_buf->chan_id_flags[chan_id][rd_ptr];
            audio_push_buf->op->get_curr_push(audio_push_buf, &curr_push_buf_addr, &curr_dma_addr, chan_id);
            if(curr_push_buf_addr != NULL){
                pci_dma_sync_single_for_cpu(chip->pci_dev, curr_dma_addr, AUDIO_PUSH_DATA_OFFSET_OF_PING_PONG, DMA_FROM_DEVICE);
                if(chan_id_flags & (1<<chan_id)){
                    src_buf = curr_push_buf_addr;
                    src_buf += (chan_id&0x3)*AUDIO_SECTION_SIZE;

                    chip_audio->op->get_audio_chan_driver(chip_audio, chan_id, &audio_chan_driver);
                    if(audio_chan_driver != NULL){
						unsigned long	flags;

                        audio_chan_push_interface = &audio_chan_driver->audio_chan_push_interface;
                        ed_tcb = &audio_chan_driver->opened_logic_chan_ed;
                        if(audio_chan_push_interface->op != NULL){
                            if(audio_chan_push_interface->op->chan_is_open(audio_chan_push_interface, audio_chan_driver) == ISIL_OK){
                                audio_chan_push_interface->op->copy_data(audio_chan_push_interface, audio_chan_driver, src_buf, AUDIO_SECTION_SIZE);
								spin_lock_irqsave(&audio_chan_driver->fsm_matrix_call_lock, flags);
                                driver_gen_deliver_event(ed_tcb, 1);
								spin_unlock_irqrestore(&audio_chan_driver->fsm_matrix_call_lock, flags);
                            }
                        }
                    }
                }
            }
        }
    }
}

/****************************************************************************************************************
* Function Name :   init_isil_audio_encode_tasklet
* Description   :   init one tasklet for copying audio bitstram
* Input         :   param1 is tasklet entry
*                   param2 is register param
* Output        :   null
* Return        :   null
****************************************************************************************************************/
void    init_isil_audio_encode_tasklet(isil_tasklet_entry_t *isil_tasklet_entry, unsigned long data)
{
    isil_tasklet_entry->op = &isil_tasklet_op;
    isil_tasklet_entry->op->init(isil_tasklet_entry, audio_encode_tasklet, data);
}

/****************************************************************************************************************
* Function Name :   close_isil_audio_encode_tasklet
* Description   :   close for audio's tasklet
* Input         :   param1 is tasklet entry
* Output        :   null
* Return        :   null
****************************************************************************************************************/
void    close_isil_audio_encode_tasklet(isil_tasklet_entry_t *isil_tasklet_entry)
{
    if(isil_tasklet_entry->op){
        isil_tasklet_entry->op->close(isil_tasklet_entry);
    }
}
#endif

